""" TaskQueueDB class is a front-end to the task queues db
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id"

import six
import random
import string

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.PrettyPrint import printDict
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.WorkloadManagementSystem.private.SharesCorrector import SharesCorrector

DEFAULT_GROUP_SHARE = 1000
TQ_MIN_SHARE = 0.001

# For checks at insertion time, and not only
singleValueDefFields = ("OwnerDN", "OwnerGroup", "Setup", "CPUTime")
multiValueDefFields = ("Sites", "GridCEs", "BannedSites", "Platforms", "SubmitPools", "JobTypes", "Tags")

# Used for matching
multiValueMatchFields = ("GridCE", "Site", "Platform", "SubmitPool", "JobType", "Tag")
bannedJobMatchFields = ("Site",)
mandatoryMatchFields = ("Setup", "CPUTime")
priorityIgnoredFields = ("Sites", "BannedSites")


def _lowerAndRemovePunctuation(s):
    if six.PY3:
        table = str.maketrans("", "", string.punctuation)  # pylint: disable=no-member
        return s.lower().translate(table)
    else:
        return s.lower().translate(None, string.punctuation)


class TaskQueueDB(DB):
    """MySQL DB of "Task Queues" """

    def __init__(self):
        DB.__init__(self, "TaskQueueDB", "WorkloadManagement/TaskQueueDB")
        self.__maxJobsInTQ = 5000
        self.__defaultCPUSegments = [
            6 * 60,
            30 * 60,
            1 * 3600,
            6 * 3600,
            12 * 3600,
            1 * 86400,
            2 * 86400,
            3 * 86400,
            4 * 86400,
            6 * 86400,
            8 * 86400,
            10 * 86400,
            int(12.5 * 86400),
        ]
        self.__maxMatchRetry = 3
        self.__jobPriorityBoundaries = (0.001, 10)
        self.__groupShares = {}
        self.__deleteTQWithDelay = DictCache(self.__deleteTQIfEmpty)
        self.__opsHelper = Operations()
        self.__ensureInsertionIsSingle = False
        self.__sharesCorrector = SharesCorrector(self.__opsHelper)
        result = self.__initializeDB()
        if not result["OK"]:
            raise Exception("Can't create tables: %s" % result["Message"])

    def enableAllTaskQueues(self):
        """Enable all Task queues"""
        return self.updateFields("tq_TaskQueues", updateDict={"Enabled": "1"})

    def findOrphanJobs(self):
        """Find jobs that are not in any task queue"""
        result = self._query("select JobID from tq_Jobs WHERE TQId not in (SELECT TQId from tq_TaskQueues)")
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    def isSharesCorrectionEnabled(self):
        return self.__getCSOption("EnableSharesCorrection", False)

    def __getCSOption(self, optionName, defValue):
        return self.__opsHelper.getValue("JobScheduling/%s" % optionName, defValue)

    def getValidPilotTypes(self):
        return self.__getCSOption("AllPilotTypes", ["private"])

    def __initializeDB(self):
        """
        Create the tables
        """
        result = self._query("show tables")
        if not result["OK"]:
            return result

        tablesInDB = [t[0] for t in result["Value"]]
        tablesToCreate = {}
        self.__tablesDesc = {}

        self.__tablesDesc["tq_TaskQueues"] = {
            "Fields": {
                "TQId": "INTEGER(11) UNSIGNED AUTO_INCREMENT NOT NULL",
                "OwnerDN": "VARCHAR(255) NOT NULL",
                "OwnerGroup": "VARCHAR(32) NOT NULL",
                "Setup": "VARCHAR(32) NOT NULL",
                "CPUTime": "BIGINT(20) UNSIGNED NOT NULL",
                "Priority": "FLOAT NOT NULL",
                "Enabled": "TINYINT(1) NOT NULL DEFAULT 0",
            },
            "PrimaryKey": "TQId",
            "Indexes": {"TQOwner": ["OwnerDN", "OwnerGroup", "Setup", "CPUTime"]},
        }

        self.__tablesDesc["tq_Jobs"] = {
            "Fields": {
                "TQId": "INTEGER(11) UNSIGNED NOT NULL",
                "JobId": "INTEGER(11) UNSIGNED NOT NULL",
                "Priority": "INTEGER UNSIGNED NOT NULL",
                "RealPriority": "FLOAT NOT NULL",
            },
            "PrimaryKey": "JobId",
            "Indexes": {"TaskIndex": ["TQId"]},
            "ForeignKeys": {"TQId": "tq_TaskQueues.TQId"},
        }

        for multiField in multiValueDefFields:
            tableName = "tq_TQTo%s" % multiField
            self.__tablesDesc[tableName] = {
                "Fields": {"TQId": "INTEGER(11) UNSIGNED NOT NULL", "Value": "VARCHAR(64) NOT NULL"},
                "PrimaryKey": ["TQId", "Value"],
                "Indexes": {"TaskIndex": ["TQId"], "%sIndex" % multiField: ["Value"]},
                "ForeignKeys": {"TQId": "tq_TaskQueues.TQId"},
            }

        for tableName in self.__tablesDesc:
            if tableName not in tablesInDB:
                tablesToCreate[tableName] = self.__tablesDesc[tableName]

        return self._createTables(tablesToCreate)

    def getGroupsInTQs(self):
        cmdSQL = "SELECT DISTINCT( OwnerGroup ) FROM `tq_TaskQueues`"
        result = self._query(cmdSQL)
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    def fitCPUTimeToSegments(self, cpuTime):
        """
        Fit the CPU time to the valid segments
        """
        maxCPUSegments = self.__getCSOption("taskQueueCPUTimeIntervals", self.__defaultCPUSegments)
        try:
            maxCPUSegments = [int(seg) for seg in maxCPUSegments]
            # Check segments in the CS
            last = 0
            for cpuS in maxCPUSegments:
                if cpuS <= last:
                    maxCPUSegments = self.__defaultCPUSegments
                    break
                last = cpuS
        except Exception:
            maxCPUSegments = self.__defaultCPUSegments
        # Map to a segment
        for cpuSegment in maxCPUSegments:
            if cpuTime <= cpuSegment:
                return cpuSegment
        return maxCPUSegments[-1]

    def _checkTaskQueueDefinition(self, tqDefDict):
        """
        Check a task queue definition dict is valid
        """

        for field in singleValueDefFields:
            if field not in tqDefDict:
                return S_ERROR("Missing mandatory field '%s' in task queue definition" % field)
            if field in ["CPUTime"]:
                if not isinstance(tqDefDict[field], six.integer_types):
                    return S_ERROR("Mandatory field %s value type is not valid: %s" % (field, type(tqDefDict[field])))
            else:
                if not isinstance(tqDefDict[field], six.string_types):
                    return S_ERROR("Mandatory field %s value type is not valid: %s" % (field, type(tqDefDict[field])))
                result = self._escapeString(tqDefDict[field])
                if not result["OK"]:
                    return result
                tqDefDict[field] = result["Value"]
        for field in multiValueDefFields:
            if field not in tqDefDict:
                continue
            if not isinstance(tqDefDict[field], (list, tuple)):
                return S_ERROR("Multi value field %s value type is not valid: %s" % (field, type(tqDefDict[field])))
            result = self._escapeValues(tqDefDict[field])
            if not result["OK"]:
                return result
            tqDefDict[field] = result["Value"]

        return S_OK(tqDefDict)

    def _checkMatchDefinition(self, tqMatchDict):
        """
        Check a task queue match dict is valid
        """

        def travelAndCheckType(value, validTypes, escapeValues=True):
            if isinstance(value, (list, tuple)):
                for subValue in value:
                    if not isinstance(subValue, validTypes):
                        return S_ERROR("List contained type %s is not valid -> %s" % (type(subValue), validTypes))
                if escapeValues:
                    return self._escapeValues(value)
                return S_OK(value)
            else:
                if not isinstance(value, validTypes):
                    return S_ERROR("Type %s is not valid -> %s" % (type(value), validTypes))
                if escapeValues:
                    return self._escapeString(value)
                return S_OK(value)

        for field in singleValueDefFields:
            if field not in tqMatchDict:
                if field in mandatoryMatchFields:
                    return S_ERROR("Missing mandatory field '%s' in match request definition" % field)
                continue
            fieldValue = tqMatchDict[field]
            if field in ["CPUTime"]:
                result = travelAndCheckType(fieldValue, six.integer_types, escapeValues=False)
            else:
                result = travelAndCheckType(fieldValue, six.string_types)
            if not result["OK"]:
                return S_ERROR("Match definition field %s failed : %s" % (field, result["Message"]))
            tqMatchDict[field] = result["Value"]
        # Check multivalue
        for multiField in multiValueMatchFields:
            for field in (multiField, "Banned%s" % multiField, "Required%s" % multiField):
                if field in tqMatchDict:
                    fieldValue = tqMatchDict[field]
                    result = travelAndCheckType(fieldValue, six.string_types)
                    if not result["OK"]:
                        return S_ERROR("Match definition field %s failed : %s" % (field, result["Message"]))
                    tqMatchDict[field] = result["Value"]

        return S_OK(tqMatchDict)

    def __createTaskQueue(self, tqDefDict, priority=1, connObj=False):
        """
        Create a task queue
          :returns: S_OK( tqId ) / S_ERROR
        """
        if not connObj:
            result = self._getConnection()
            if not result["OK"]:
                return S_ERROR("Can't create task queue: %s" % result["Message"])
            connObj = result["Value"]
        tqDefDict["CPUTime"] = self.fitCPUTimeToSegments(tqDefDict["CPUTime"])
        sqlSingleFields = ["TQId", "Priority"]
        sqlValues = ["0", str(priority)]
        for field in singleValueDefFields:
            sqlSingleFields.append(field)
            sqlValues.append(tqDefDict[field])
        # Insert the TQ Disabled
        sqlSingleFields.append("Enabled")
        sqlValues.append("0")
        cmd = "INSERT INTO tq_TaskQueues ( %s ) VALUES ( %s )" % (
            ", ".join(sqlSingleFields),
            ", ".join([str(v) for v in sqlValues]),
        )
        result = self._update(cmd, conn=connObj)
        if not result["OK"]:
            self.log.error("Can't insert TQ in DB", result["Value"])
            return result
        if "lastRowId" in result:
            tqId = result["lastRowId"]
        else:
            result = self._query("SELECT LAST_INSERT_ID()", conn=connObj)
            if not result["OK"]:
                self.cleanOrphanedTaskQueues(connObj=connObj)
                return S_ERROR("Can't determine task queue id after insertion")
            tqId = result["Value"][0][0]
        for field in multiValueDefFields:
            if field not in tqDefDict:
                continue
            values = List.uniqueElements([value for value in tqDefDict[field] if value.strip()])
            if not values:
                continue
            cmd = "INSERT INTO `tq_TQTo%s` ( TQId, Value ) VALUES " % field
            cmd += ", ".join(["( %s, %s )" % (tqId, str(value)) for value in values])
            result = self._update(cmd, conn=connObj)
            if not result["OK"]:
                self.log.error("Failed to insert condition", "%s : %s" % field, result["Message"])
                self.cleanOrphanedTaskQueues(connObj=connObj)
                return S_ERROR("Can't insert values %s for field %s: %s" % (str(values), field, result["Message"]))
        self.log.info("Created TQ", tqId)
        return S_OK(tqId)

    def cleanOrphanedTaskQueues(self, connObj=False):
        """
        Delete all empty task queues
        """
        self.log.info("Cleaning orphaned TQs")
        sq = "SELECT TQId FROM `tq_TaskQueues` WHERE Enabled >= 1 AND TQId not in ( SELECT DISTINCT TQId from `tq_Jobs` )"
        result = self._query(sq, conn=connObj)
        if not result["OK"]:
            return result
        orphanedTQs = result["Value"]
        if not orphanedTQs:
            return S_OK()
        orphanedTQs = [str(otq[0]) for otq in orphanedTQs]

        for mvField in multiValueDefFields:
            result = self._update(
                "DELETE FROM `tq_TQTo%s` WHERE TQId in ( %s )" % (mvField, ",".join(orphanedTQs)), conn=connObj
            )
            if not result["OK"]:
                return result

        result = self._update("DELETE FROM `tq_TaskQueues` WHERE TQId in ( %s )" % ",".join(orphanedTQs), conn=connObj)
        if not result["OK"]:
            return result
        return S_OK()

    def __setTaskQueueEnabled(self, tqId, enabled=True, connObj=False):
        if enabled:
            enabled = "+ 1"
        else:
            enabled = "- 1"
        upSQL = "UPDATE `tq_TaskQueues` SET Enabled = Enabled %s WHERE TQId=%d" % (enabled, tqId)
        result = self._update(upSQL, conn=connObj)
        if not result["OK"]:
            self.log.error("Error setting TQ state", "TQ %s State %s: %s" % (tqId, enabled, result["Message"]))
            return result
        updated = result["Value"] > 0
        if updated:
            self.log.verbose("Set enabled for TQ", "(%s for TQ %s)" % (enabled, tqId))
        return S_OK(updated)

    def __hackJobPriority(self, jobPriority):
        jobPriority = min(max(int(jobPriority), self.__jobPriorityBoundaries[0]), self.__jobPriorityBoundaries[1])
        if jobPriority == self.__jobPriorityBoundaries[0]:
            return 10 ** (-5)
        if jobPriority == self.__jobPriorityBoundaries[1]:
            return 10 ** 6
        return jobPriority

    def insertJob(self, jobId, tqDefDict, jobPriority, skipTQDefCheck=False):
        """Insert a job in a task queue (creating one if it doesn't exit)

        :param int jobId: job ID
        :param dict tqDefDict: dict for TQ definition
        :param int jobPriority: integer that defines the job priority

        :returns: S_OK() / S_ERROR
        """
        try:
            int(jobId)
        except ValueError:
            return S_ERROR("JobId is not a number!")
        retVal = self._getConnection()
        if not retVal["OK"]:
            return S_ERROR("Can't insert job: %s" % retVal["Message"])
        connObj = retVal["Value"]
        if not skipTQDefCheck:
            tqDefDict = dict(tqDefDict)
            retVal = self._checkTaskQueueDefinition(tqDefDict)
            if not retVal["OK"]:
                self.log.error("TQ definition check failed", retVal["Message"])
                return retVal
            tqDefDict = retVal["Value"]
        tqDefDict["CPUTime"] = self.fitCPUTimeToSegments(tqDefDict["CPUTime"])
        self.log.info("Inserting job with requirements", "(%s : %s)" % (jobId, printDict(tqDefDict)))
        retVal = self.__findAndDisableTaskQueue(tqDefDict, skipDefinitionCheck=True, connObj=connObj)
        if not retVal["OK"]:
            return retVal
        tqInfo = retVal["Value"]
        newTQ = False
        if not tqInfo["found"]:
            self.log.info("Creating a TQ for job", jobId)
            retVal = self.__createTaskQueue(tqDefDict, 1, connObj=connObj)
            if not retVal["OK"]:
                return retVal
            tqId = retVal["Value"]
            newTQ = True
        else:
            tqId = tqInfo["tqId"]
            self.log.info("Found TQ for job requirements", "(%s : %s)" % (tqId, jobId))
        try:
            result = self.__insertJobInTaskQueue(jobId, tqId, int(jobPriority), checkTQExists=False, connObj=connObj)
            if not result["OK"]:
                self.log.error("Error inserting job in TQ", "Job %s TQ %s: %s" % (jobId, tqId, result["Message"]))
                return result
            if newTQ:
                self.recalculateTQSharesForEntity(tqDefDict["OwnerDN"], tqDefDict["OwnerGroup"], connObj=connObj)
        finally:
            self.__setTaskQueueEnabled(tqId, True)
        return S_OK()

    def __insertJobInTaskQueue(self, jobId, tqId, jobPriority, checkTQExists=True, connObj=False):
        """Insert a job in a given task queue

        :param int jobId: job ID
        :param dict tqDefDict: dict for TQ definition
        :param int jobPriority: integer that defines the job priority

        :returns: S_OK() / S_ERROR
        """
        self.log.info("Inserting job in TQ with priority", "(%s : %s : %s)" % (jobId, tqId, jobPriority))
        if not connObj:
            result = self._getConnection()
            if not result["OK"]:
                return S_ERROR("Can't insert job: %s" % result["Message"])
            connObj = result["Value"]
        if checkTQExists:
            result = self._query("SELECT tqId FROM `tq_TaskQueues` WHERE TQId = %s" % tqId, conn=connObj)
            if not result["OK"] or not result["Value"]:
                return S_OK("Can't find task queue with id %s: %s" % (tqId, result["Message"]))
        hackedPriority = self.__hackJobPriority(jobPriority)
        result = self._update(
            "INSERT INTO tq_Jobs ( TQId, JobId, Priority, RealPriority ) \
                            VALUES ( %s, %s, %s, %f ) ON DUPLICATE KEY UPDATE TQId = %s, \
                            Priority = %s, RealPriority = %f"
            % (tqId, jobId, jobPriority, hackedPriority, tqId, jobPriority, hackedPriority),
            conn=connObj,
        )
        if not result["OK"]:
            return result
        return S_OK()

    def __generateTQFindSQL(self, tqDefDict, skipDefinitionCheck=False):
        """
        Generate the SQL to find a task queue that has exactly the given requirements

        :param dict tqDefDict: dict for TQ definition
        :returns: S_OK() / S_ERROR
        """
        if not skipDefinitionCheck:
            tqDefDict = dict(tqDefDict)
            result = self._checkTaskQueueDefinition(tqDefDict)
            if not result["OK"]:
                return result
            tqDefDict = result["Value"]

        sqlCondList = []
        for field in singleValueDefFields:
            sqlCondList.append("`tq_TaskQueues`.%s = %s" % (field, tqDefDict[field]))
        # MAGIC SUBQUERIES TO ENSURE STRICT MATCH
        for field in multiValueDefFields:
            tableName = "`tq_TQTo%s`" % field
            if field in tqDefDict and tqDefDict[field]:
                firstQuery = (
                    "SELECT COUNT(%s.Value) \
                      FROM %s \
                      WHERE %s.TQId = `tq_TaskQueues`.TQId"
                    % (tableName, tableName, tableName)
                )
                grouping = "GROUP BY %s.TQId" % tableName
                valuesList = List.uniqueElements([value.strip() for value in tqDefDict[field] if value.strip()])
                numValues = len(valuesList)
                secondQuery = "%s AND %s.Value in (%s)" % (
                    firstQuery,
                    tableName,
                    ",".join(["%s" % str(value) for value in valuesList]),
                )
                sqlCondList.append("%s = (%s %s)" % (numValues, firstQuery, grouping))
                sqlCondList.append("%s = (%s %s)" % (numValues, secondQuery, grouping))
            else:
                sqlCondList.append(
                    "`tq_TaskQueues`.TQId not in ( SELECT DISTINCT %s.TQId from %s )" % (tableName, tableName)
                )
        # END MAGIC: That was easy ;)
        return S_OK(" AND ".join(sqlCondList))

    def __findAndDisableTaskQueue(self, tqDefDict, skipDefinitionCheck=False, retries=10, connObj=False):
        """Disable and find TQ

        :param dict tqDefDict: dict for TQ definition
        :returns: S_OK() / S_ERROR
        """
        for _ in range(retries):
            result = self.__findSmallestTaskQueue(tqDefDict, skipDefinitionCheck=skipDefinitionCheck, connObj=connObj)
            if not result["OK"]:
                return result
            data = result["Value"]
            if not data["found"]:
                return result
            if data["enabled"] < 1:
                self.log.debug("TaskQueue {tqId} seems to be already disabled ({enabled})".format(**data))
            result = self.__setTaskQueueEnabled(data["tqId"], False)
            if result["OK"]:
                return S_OK(data)
        return S_ERROR("Could not disable TQ")

    def __findSmallestTaskQueue(self, tqDefDict, skipDefinitionCheck=False, connObj=False):
        """
        Find a task queue that has at least the given requirements

        :param dict tqDefDict: dict for TQ definition
        :returns: S_OK() / S_ERROR
        """
        result = self.__generateTQFindSQL(tqDefDict, skipDefinitionCheck=skipDefinitionCheck)
        if not result["OK"]:
            return result

        sqlCmd = "SELECT COUNT( `tq_Jobs`.JobID ), `tq_TaskQueues`.TQId, `tq_TaskQueues`.Enabled \
FROM `tq_TaskQueues`, `tq_Jobs`"
        sqlCmd = (
            "%s WHERE `tq_TaskQueues`.TQId = `tq_Jobs`.TQId AND %s GROUP BY `tq_Jobs`.TQId \
ORDER BY COUNT( `tq_Jobs`.JobID ) ASC"
            % (sqlCmd, result["Value"])
        )
        result = self._query(sqlCmd, conn=connObj)
        if not result["OK"]:
            self.log.error("Can't find task queue", result["Message"])
            return result
        data = result["Value"]
        if not data or data[0][0] >= self.__maxJobsInTQ:
            return S_OK({"found": False})
        return S_OK({"found": True, "tqId": data[0][1], "enabled": data[0][2], "jobs": data[0][0]})

    def matchAndGetJob(self, tqMatchDict, numJobsPerTry=50, numQueuesPerTry=10, negativeCond=None):
        """Match a job based on requirements

        :param dict tqDefDict: dict for TQ definition
        :returns: S_OK() / S_ERROR
        """
        if negativeCond is None:
            negativeCond = {}
        # Make a copy to avoid modification of original if escaping needs to be done
        tqMatchDict = dict(tqMatchDict)
        retVal = self._checkMatchDefinition(tqMatchDict)
        if not retVal["OK"]:
            self.log.error("TQ match request check failed", retVal["Message"])
            return retVal
        retVal = self._getConnection()
        if not retVal["OK"]:
            return S_ERROR("Can't connect to DB: %s" % retVal["Message"])
        connObj = retVal["Value"]
        preJobSQL = "SELECT `tq_Jobs`.JobId, `tq_Jobs`.TQId \
FROM `tq_Jobs` WHERE `tq_Jobs`.TQId = %s AND `tq_Jobs`.Priority = %s"
        prioSQL = "SELECT `tq_Jobs`.Priority FROM `tq_Jobs` \
WHERE `tq_Jobs`.TQId = %s ORDER BY RAND() / `tq_Jobs`.RealPriority ASC LIMIT 1"
        postJobSQL = " ORDER BY `tq_Jobs`.JobId ASC LIMIT %s" % numJobsPerTry
        for _ in range(self.__maxMatchRetry):
            noJobsFound = False
            if "JobID" in tqMatchDict:
                # A certain JobID is required by the resource, so all TQ are to be considered
                retVal = self.matchAndGetTaskQueue(
                    tqMatchDict, numQueuesToGet=0, skipMatchDictDef=True, connObj=connObj
                )
                preJobSQL = "%s AND `tq_Jobs`.JobId = %s " % (preJobSQL, tqMatchDict["JobID"])
            else:
                retVal = self.matchAndGetTaskQueue(
                    tqMatchDict,
                    numQueuesToGet=numQueuesPerTry,
                    skipMatchDictDef=True,
                    negativeCond=negativeCond,
                    connObj=connObj,
                )
            if not retVal["OK"]:
                return retVal
            tqList = retVal["Value"]
            if not tqList:
                self.log.info("No TQ matches requirements")
                return S_OK({"matchFound": False, "tqMatch": tqMatchDict})
            for tqId, tqOwnerDN, tqOwnerGroup in tqList:
                self.log.verbose("Trying to extract jobs from TQ", tqId)
                retVal = self._query(prioSQL % tqId, conn=connObj)
                if not retVal["OK"]:
                    return S_ERROR("Can't retrieve winning priority for matching job: %s" % retVal["Message"])
                if not retVal["Value"]:
                    noJobsFound = True
                    continue
                prio = retVal["Value"][0][0]
                retVal = self._query("%s %s" % (preJobSQL % (tqId, prio), postJobSQL), conn=connObj)
                if not retVal["OK"]:
                    return S_ERROR("Can't begin transaction for matching job: %s" % retVal["Message"])
                jobTQList = [(row[0], row[1]) for row in retVal["Value"]]
                if not jobTQList:
                    self.log.info("Task queue seems to be empty, triggering a cleaning of", tqId)
                    self.__deleteTQWithDelay.add(tqId, 300, (tqId, tqOwnerDN, tqOwnerGroup))
                while jobTQList:
                    jobId, tqId = jobTQList.pop(random.randint(0, len(jobTQList) - 1))
                    self.log.verbose("Trying to extract job from TQ", "%s : %s" % (jobId, tqId))
                    retVal = self.deleteJob(jobId, connObj=connObj)
                    if not retVal["OK"]:
                        msgFix = "Could not take job"
                        msgVar = " %s out from the TQ %s: %s" % (jobId, tqId, retVal["Message"])
                        self.log.error(msgFix, msgVar)
                        return S_ERROR(msgFix + msgVar)
                    if retVal["Value"]:
                        self.log.info("Extracted job with prio from TQ", "(%s : %s : %s)" % (jobId, prio, tqId))
                        return S_OK({"matchFound": True, "jobId": jobId, "taskQueueId": tqId, "tqMatch": tqMatchDict})
                self.log.info("No jobs could be extracted from TQ", tqId)
        if noJobsFound:
            return S_OK({"matchFound": False, "tqMatch": tqMatchDict})

        self.log.info("Could not find a match after %s match retries" % self.__maxMatchRetry)
        return S_ERROR("Could not find a match after %s match retries" % self.__maxMatchRetry)

    def matchAndGetTaskQueue(
        self, tqMatchDict, numQueuesToGet=1, skipMatchDictDef=False, negativeCond=None, connObj=False
    ):
        """Get a queue that matches the requirements"""
        if negativeCond is None:
            negativeCond = {}
        # Make a copy to avoid modification of original if escaping needs to be done
        tqMatchDict = dict(tqMatchDict)
        if not skipMatchDictDef:
            retVal = self._checkMatchDefinition(tqMatchDict)
            if not retVal["OK"]:
                return retVal
        retVal = self.__generateTQMatchSQL(tqMatchDict, numQueuesToGet=numQueuesToGet, negativeCond=negativeCond)
        if not retVal["OK"]:
            return retVal
        matchSQL = retVal["Value"]
        retVal = self._query(matchSQL, conn=connObj)
        if not retVal["OK"]:
            return retVal
        return S_OK([(row[0], row[1], row[2]) for row in retVal["Value"]])

    @staticmethod
    def __generateSQLSubCond(sqlString, value, boolOp="OR"):
        if not isinstance(value, (list, tuple)):
            return sqlString % str(value).strip()
        sqlORList = []
        for v in value:
            sqlORList.append(sqlString % str(v).strip())
        return "( %s )" % (" %s " % boolOp).join(sqlORList)

    def __generateNotSQL(self, negativeCond):
        """Generate negative conditions
        Can be a list of dicts or a dict:
         - list of dicts will be  OR of conditional dicts
         - dicts will be normal conditional dict ( kay1 in ( v1, v2, ... ) AND key2 in ( v3, v4, ... ) )
        """
        if isinstance(negativeCond, (list, tuple)):
            sqlCond = []
            for cD in negativeCond:
                sqlCond.append(self.__generateNotDictSQL(cD))
            return " ( %s )" % " OR  ".join(sqlCond)
        elif isinstance(negativeCond, dict):
            return self.__generateNotDictSQL(negativeCond)
        raise RuntimeError(
            "negativeCond has to be either a list or a dict or a tuple, and it's %s" % type(negativeCond)
        )

    def __generateNotDictSQL(self, negativeCond):
        """Generate the negative sql condition from a standard condition dict
        not ( cond1 and cond2 ) = ( not cond1 or not cond 2 )
        For instance: { 'Site': 'S1', 'JobType': [ 'T1', 'T2' ] }
          ( not 'S1' in Sites or ( not 'T1' in JobType and not 'T2' in JobType ) )
          S2 T1 -> not False or ( not True and not False ) -> True or ... -> True -> Eligible
          S1 T3 -> not True or ( not False and not False ) -> False or (True and True ) -> True -> Eligible
          S1 T1 -> not True or ( not True and not False ) -> False or ( False and True ) -> False -> Nop
        """
        condList = []
        for field in negativeCond:
            if field in multiValueMatchFields:
                fullTableN = "`tq_TQTo%ss`" % field
                valList = negativeCond[field]
                if not isinstance(valList, (list, tuple)):
                    valList = (valList,)
                subList = []
                for value in valList:
                    value = self._escapeString(value)["Value"]
                    sql = "%s NOT IN ( SELECT %s.Value FROM %s WHERE %s.TQId = tq.TQId )" % (
                        value,
                        fullTableN,
                        fullTableN,
                        fullTableN,
                    )
                    subList.append(sql)
                condList.append("( %s )" % " AND ".join(subList))
            elif field in singleValueDefFields:
                for value in negativeCond[field]:
                    value = self._escapeString(value)["Value"]
                    sql = "%s != tq.%s " % (value, field)
                    condList.append(sql)
        return "( %s )" % " OR ".join(condList)

    @staticmethod
    def __generateTablesName(sqlTables, field):
        fullTableName = "tq_TQTo%ss" % field
        if fullTableName not in sqlTables:
            tableN = field.lower()
            sqlTables[fullTableName] = tableN
            return (
                tableN,
                "`%s`" % fullTableName,
            )
        return sqlTables[fullTableName], "`%s`" % fullTableName

    def __generateTQMatchSQL(self, tqMatchDict, numQueuesToGet=1, negativeCond=None):
        """
        Generate the SQL needed to match a task queue
        """
        self.log.debug(tqMatchDict)

        if negativeCond is None:
            negativeCond = {}
        # Only enabled TQs
        sqlCondList = []
        sqlTables = {"tq_TaskQueues": "tq"}
        # If OwnerDN and OwnerGroup are defined only use those combinations that make sense
        if "OwnerDN" in tqMatchDict and "OwnerGroup" in tqMatchDict:
            groups = tqMatchDict["OwnerGroup"]
            if not isinstance(groups, (list, tuple)):
                groups = [groups]
            dns = tqMatchDict["OwnerDN"]
            if not isinstance(dns, (list, tuple)):
                dns = [dns]
            ownerConds = []
            for group in groups:
                if Properties.JOB_SHARING in Registry.getPropertiesForGroup(group.replace('"', "")):
                    ownerConds.append("tq.OwnerGroup = %s" % group)
                else:
                    for dn in dns:
                        ownerConds.append("( tq.OwnerDN = %s AND tq.OwnerGroup = %s )" % (dn, group))
            sqlCondList.append(" OR ".join(ownerConds))
        else:
            # If not both are defined, just add the ones that are defined
            for field in ("OwnerGroup", "OwnerDN"):
                if field in tqMatchDict:
                    sqlCondList.append(self.__generateSQLSubCond("tq.%s = %%s" % field, tqMatchDict[field]))
        # Type of single value conditions
        for field in ("CPUTime", "Setup"):
            if field in tqMatchDict:
                if field == "CPUTime":
                    sqlCondList.append(self.__generateSQLSubCond("tq.%s <= %%s" % field, tqMatchDict[field]))
                else:
                    sqlCondList.append(self.__generateSQLSubCond("tq.%s = %%s" % field, tqMatchDict[field]))

        tag_fv = []

        # Match multi value fields
        for field in multiValueMatchFields:
            self.log.debug("Evaluating field %s" % field)
            # It has to be %ss , with an 's' at the end because the columns names
            # are plural and match options are singular

            # Just treating the (not so) special case of no Tag, No RequiredTag
            if "Tag" not in tqMatchDict and "RequiredTag" not in tqMatchDict:
                tqMatchDict["Tag"] = []

            if field in tqMatchDict:
                self.log.debug("Evaluating %s with value %s" % (field, tqMatchDict[field]))

                _, fullTableN = self.__generateTablesName(sqlTables, field)

                sqlMultiCondList = []
                csql = None

                # Now evaluating Tags
                if field == "Tag":
                    tag_fv = tqMatchDict.get("Tag")
                    self.log.debug("Evaluating tag %s of type %s" % (tag_fv, type(tag_fv)))
                    if isinstance(tag_fv, str):
                        tag_fv = [tag_fv]

                    # Is there something to consider?
                    if any(_lowerAndRemovePunctuation(fvx) == "any" for fvx in tag_fv):
                        continue
                    else:
                        sqlMultiCondList.append(self.__generateTagSQLSubCond(fullTableN, tag_fv))

                # Now evaluating everything that is not tags
                else:
                    fv = tqMatchDict.get(field)
                    self.log.debug("Evaluating field %s of type %s" % (field, type(fv)))

                    # Is there something to consider?
                    if not fv:
                        continue
                    if isinstance(fv, str) and _lowerAndRemovePunctuation(fv) == "any":
                        continue
                    if isinstance(fv, list) and any(_lowerAndRemovePunctuation(fvx) == "any" for fvx in fv):
                        continue
                    # if field != 'GridCE' or 'Site' in tqMatchDict:
                    # Jobs for masked sites can be matched if they specified a GridCE
                    # Site is removed from tqMatchDict if the Site is mask. In this case we want
                    # that the GridCE matches explicitly so the COUNT can not be 0. In this case we skip this
                    # condition
                    sqlMultiCondList.append(
                        "( SELECT COUNT(%s.Value) FROM %s WHERE %s.TQId = tq.TQId ) = 0"
                        % (fullTableN, fullTableN, fullTableN)
                    )
                    sqlMultiCondList.append(
                        self.__generateSQLSubCond(
                            "%%s IN ( SELECT %s.Value \
                                                            FROM %s \
                                                            WHERE %s.TQId = tq.TQId )"
                            % (fullTableN, fullTableN, fullTableN),
                            tqMatchDict.get(field),
                        )
                    )

                sqlCondList.append("( %s )" % " OR ".join(sqlMultiCondList))

                # In case of Site, check it's not in job banned sites
                if field in bannedJobMatchFields:
                    fullTableN = "`tq_TQToBanned%ss`" % field
                    csql = self.__generateSQLSubCond(
                        "%%s not in ( SELECT %s.Value \
                                                          FROM %s \
                                                          WHERE %s.TQId = tq.TQId )"
                        % (fullTableN, fullTableN, fullTableN),
                        tqMatchDict[field],
                        boolOp="OR",
                    )
                    sqlCondList.append(csql)

        # Add possibly RequiredTag conditions
        rtag_fv = tqMatchDict.get("RequiredTag", [])
        if isinstance(rtag_fv, str):
            rtag_fv = [rtag_fv]

        # Is there something to consider?
        if not rtag_fv or any(_lowerAndRemovePunctuation(fv) == "any" for fv in rtag_fv):
            pass
        elif not set(rtag_fv).issubset(set(tag_fv)):
            return S_ERROR("Wrong conditions")
        else:
            self.log.debug("Evaluating RequiredTag %s" % rtag_fv)
            sqlCondList.append(self.__generateRequiredTagSQLSubCond("`tq_TQToTags`", rtag_fv))

        # Add possibly Resource banning conditions
        for field in multiValueMatchFields:
            bannedField = "Banned%s" % field

            # Is there something to consider?
            b_fv = tqMatchDict.get(bannedField)
            if (
                not b_fv
                or isinstance(b_fv, str)
                and _lowerAndRemovePunctuation(b_fv) == "any"
                or isinstance(b_fv, list)
                and any(_lowerAndRemovePunctuation(fvx) == "any" for fvx in b_fv)
            ):
                continue

            fullTableN = "`tq_TQTo%ss`" % field

            sqlCondList.append(
                self.__generateSQLSubCond(
                    "%%s not in ( SELECT %s.Value \
                                                      FROM %s \
                                                      WHERE %s.TQId = tq.TQId )"
                    % (fullTableN, fullTableN, fullTableN),
                    b_fv,
                    boolOp="OR",
                )
            )

        # Add extra negative conditions
        if negativeCond:
            sqlCondList.append(self.__generateNotSQL(negativeCond))

        # Generate the final query string
        tqSqlCmd = "SELECT tq.TQId, tq.OwnerDN, tq.OwnerGroup FROM `tq_TaskQueues` tq WHERE %s" % (
            " AND ".join(sqlCondList)
        )

        # Apply priorities
        tqSqlCmd = "%s ORDER BY RAND() / tq.Priority ASC" % tqSqlCmd

        # Do we want a limit?
        if numQueuesToGet:
            tqSqlCmd = "%s LIMIT %s" % (tqSqlCmd, numQueuesToGet)
        return S_OK(tqSqlCmd)

    @staticmethod
    def __generateTagSQLSubCond(tableName, tagMatchList):
        """Generate SQL condition where ALL the specified multiValue requirements must be
        present in the matching resource list
        """
        sql1 = "SELECT COUNT(%s.Value) FROM %s WHERE %s.TQId=tq.TQId" % (tableName, tableName, tableName)
        if not tagMatchList:
            sql2 = sql1 + " AND %s.Value=''" % tableName
        else:
            if isinstance(tagMatchList, (list, tuple)):
                sql2 = sql1 + " AND %s.Value in ( %s )" % (tableName, ",".join(["%s" % v for v in tagMatchList]))
            else:
                sql2 = sql1 + " AND %s.Value=%s" % (tableName, tagMatchList)
        sql = "( " + sql1 + " ) = (" + sql2 + " )"
        return sql

    @staticmethod
    def __generateRequiredTagSQLSubCond(tableName, tagMatchList):
        """Generate SQL condition where the TQ corresponds to the requirements
        of the resource
        """
        sql = "SELECT COUNT(%s.Value) FROM %s WHERE %s.TQId=tq.TQId" % (tableName, tableName, tableName)
        if isinstance(tagMatchList, (list, tuple)):
            sql = sql + " AND %s.Value in ( %s )" % (tableName, ",".join(["%s" % v for v in tagMatchList]))
            nTags = len(tagMatchList)
        else:
            sql = sql + " AND %s.Value=%s" % (tableName, tagMatchList)
            nTags = 1
        sql = "( %s ) = %s" % (sql, nTags)
        return sql

    def deleteJob(self, jobId, connObj=False):
        """
        Delete a job from the task queues
        Return S_OK( True/False ) / S_ERROR
        """
        if not connObj:
            retVal = self._getConnection()
            if not retVal["OK"]:
                return S_ERROR("Can't delete job: %s" % retVal["Message"])
            connObj = retVal["Value"]
        retVal = self._query(
            "SELECT t.TQId, t.OwnerDN, t.OwnerGroup \
FROM `tq_TaskQueues` t, `tq_Jobs` j \
WHERE j.JobId = %s AND t.TQId = j.TQId"
            % jobId,
            conn=connObj,
        )
        if not retVal["OK"]:
            return S_ERROR("Could not get job from task queue %s: %s" % (jobId, retVal["Message"]))
        data = retVal["Value"]
        if not data:
            return S_OK(False)
        tqId, tqOwnerDN, tqOwnerGroup = data[0]
        self.log.verbose("Deleting job", jobId)
        retVal = self._update("DELETE FROM `tq_Jobs` WHERE JobId = %s" % jobId, conn=connObj)
        if not retVal["OK"]:
            return S_ERROR("Could not delete job from task queue %s: %s" % (jobId, retVal["Message"]))
        if retVal["Value"] == 0:
            # No job deleted
            return S_OK(False)
        # Always return S_OK() because job has already been taken out from the TQ
        self.__deleteTQWithDelay.add(tqId, 300, (tqId, tqOwnerDN, tqOwnerGroup))
        return S_OK(True)

    def getTaskQueueForJob(self, jobId, connObj=False):
        """
        Return TaskQueue for a given Job
        Return S_OK( [TaskQueueID] ) / S_ERROR
        """
        if not connObj:
            retVal = self._getConnection()
            if not retVal["OK"]:
                return S_ERROR("Can't get TQ for job: %s" % retVal["Message"])
            connObj = retVal["Value"]

        retVal = self._query("SELECT TQId FROM `tq_Jobs` WHERE JobId = %s" % jobId, conn=connObj)

        if not retVal["OK"]:
            return retVal

        if not retVal["Value"]:
            return S_ERROR("Not in TaskQueues")

        return S_OK(retVal["Value"][0][0])

    def getTaskQueueForJobs(self, jobIDs, connObj=False):
        """
        Return TaskQueues for a given list of Jobs
        """
        if not connObj:
            retVal = self._getConnection()
            if not retVal["OK"]:
                self.log.error("Can't get TQs for a job list", retVal["Message"])
                return retVal
            connObj = retVal["Value"]

        cmd = "SELECT JobId,TQId FROM `tq_Jobs` WHERE JobId IN (%s) " % ",".join(str(x) for x in jobIDs)
        retVal = self._query(cmd, conn=connObj)

        if not retVal["OK"]:
            return retVal

        if not retVal["Value"]:
            return S_ERROR("Not in TaskQueues")

        resultDict = {}
        for jobID, tqID in retVal["Value"]:
            resultDict[int(jobID)] = int(tqID)

        return S_OK(resultDict)

    def __getOwnerForTaskQueue(self, tqId, connObj=False):
        retVal = self._query("SELECT OwnerDN, OwnerGroup from `tq_TaskQueues` WHERE TQId=%s" % tqId, conn=connObj)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        if not data:
            return S_OK(False)
        return S_OK(retVal["Value"][0])

    def __deleteTQIfEmpty(self, args):
        (tqId, tqOwnerDN, tqOwnerGroup) = args
        retries = 3
        while retries:
            retries -= 1
            result = self.deleteTaskQueueIfEmpty(tqId, tqOwnerDN, tqOwnerGroup)
            if result["OK"]:
                return
        self.log.error("Could not delete TQ", "%s: %s" % (tqId, result["Message"]))

    def deleteTaskQueueIfEmpty(self, tqId, tqOwnerDN=False, tqOwnerGroup=False, connObj=False):
        """
        Try to delete a task queue if its empty
        """
        if not connObj:
            retVal = self._getConnection()
            if not retVal["OK"]:
                self.log.error("Can't insert job", retVal["Message"])
                return retVal
            connObj = retVal["Value"]
        if not tqOwnerDN or not tqOwnerGroup:
            retVal = self.__getOwnerForTaskQueue(tqId, connObj=connObj)
            if not retVal["OK"]:
                return retVal
            data = retVal["Value"]
            if not data:
                return S_OK(False)
            tqOwnerDN, tqOwnerGroup = data

        sqlCmd = "SELECT TQId FROM `tq_TaskQueues` WHERE Enabled >= 1 AND `tq_TaskQueues`.TQId = %s " % tqId
        sqlCmd += "AND `tq_TaskQueues`.TQId not in ( SELECT DISTINCT TQId from `tq_Jobs` )"
        retVal = self._query(sqlCmd, conn=connObj)
        if not retVal["OK"]:
            self.log.error("Could not select task queue", "%s : %s" % tqId, retVal["Message"])
            return retVal
        tqToDel = retVal["Value"]

        if tqToDel:
            for mvField in multiValueDefFields:
                retVal = self._update("DELETE FROM `tq_TQTo%s` WHERE TQId = %s" % (mvField, tqId), conn=connObj)
                if not retVal["OK"]:
                    return retVal
            retVal = self._update("DELETE FROM `tq_TaskQueues` WHERE TQId = %s" % tqId, conn=connObj)
            if not retVal["OK"]:
                return retVal
            self.recalculateTQSharesForEntity(tqOwnerDN, tqOwnerGroup, connObj=connObj)
            self.log.info("Deleted empty and enabled TQ", tqId)
            return S_OK()
        return S_OK(False)

    def deleteTaskQueue(self, tqId, tqOwnerDN=False, tqOwnerGroup=False, connObj=False):
        """
        Try to delete a task queue even if it has jobs
        """
        self.log.info("Deleting TQ", tqId)
        if not connObj:
            retVal = self._getConnection()
            if not retVal["OK"]:
                return S_ERROR("Can't insert job: %s" % retVal["Message"])
            connObj = retVal["Value"]
        if not tqOwnerDN or not tqOwnerGroup:
            retVal = self.__getOwnerForTaskQueue(tqId, connObj=connObj)
            if not retVal["OK"]:
                return retVal
            data = retVal["Value"]
            if not data:
                return S_OK(False)
            tqOwnerDN, tqOwnerGroup = data
        sqlCmd = "DELETE FROM `tq_TaskQueues` WHERE `tq_TaskQueues`.TQId = %s" % tqId
        retVal = self._update(sqlCmd, conn=connObj)
        if not retVal["OK"]:
            return S_ERROR("Could not delete task queue %s: %s" % (tqId, retVal["Message"]))
        delTQ = retVal["Value"]
        sqlCmd = "DELETE FROM `tq_Jobs` WHERE `tq_Jobs`.TQId = %s" % tqId
        retVal = self._update(sqlCmd, conn=connObj)
        if not retVal["OK"]:
            return S_ERROR("Could not delete task queue %s: %s" % (tqId, retVal["Message"]))
        for field in multiValueDefFields:
            retVal = self._update("DELETE FROM `tq_TQTo%s` WHERE TQId = %s" % (field, tqId), conn=connObj)
            if not retVal["OK"]:
                return retVal
        if delTQ > 0:
            self.recalculateTQSharesForEntity(tqOwnerDN, tqOwnerGroup, connObj=connObj)
            return S_OK(True)
        return S_OK(False)

    def getMatchingTaskQueues(self, tqMatchDict, negativeCond=False):
        """Get the info of the task queues that match a resource"""
        result = self.matchAndGetTaskQueue(tqMatchDict, numQueuesToGet=0, negativeCond=negativeCond)
        if not result["OK"]:
            return result
        return self.retrieveTaskQueues([tqTuple[0] for tqTuple in result["Value"]])

    def getNumTaskQueues(self):
        """
        Get the number of task queues in the system
        """
        sqlCmd = "SELECT COUNT( TQId ) FROM `tq_TaskQueues`"
        retVal = self._query(sqlCmd)
        if not retVal["OK"]:
            return retVal
        return S_OK(retVal["Value"][0][0])

    def retrieveTaskQueues(self, tqIdList=None):
        """
        Get all the task queues
        """
        sqlSelectEntries = ["`tq_TaskQueues`.TQId", "`tq_TaskQueues`.Priority", "COUNT( `tq_Jobs`.TQId )"]
        sqlGroupEntries = ["`tq_TaskQueues`.TQId", "`tq_TaskQueues`.Priority"]
        for field in singleValueDefFields:
            sqlSelectEntries.append("`tq_TaskQueues`.%s" % field)
            sqlGroupEntries.append("`tq_TaskQueues`.%s" % field)
        sqlCmd = "SELECT %s FROM `tq_TaskQueues`, `tq_Jobs`" % ", ".join(sqlSelectEntries)
        sqlTQCond = ""
        if tqIdList is not None:
            if not tqIdList:
                # Empty list => Fast-track no matches
                return S_OK({})
            else:
                sqlTQCond += " AND `tq_TaskQueues`.TQId in ( %s )" % ", ".join([str(id_) for id_ in tqIdList])
        sqlCmd = "%s WHERE `tq_TaskQueues`.TQId = `tq_Jobs`.TQId %s GROUP BY %s" % (
            sqlCmd,
            sqlTQCond,
            ", ".join(sqlGroupEntries),
        )
        retVal = self._query(sqlCmd)
        if not retVal["OK"]:
            self.log.error("Can't retrieve task queues info", retVal["Message"])
            return retVal
        tqData = {}
        for record in retVal["Value"]:
            tqId = record[0]
            tqData[tqId] = {"Priority": record[1], "Jobs": record[2]}
            record = record[3:]
            for iP, _ in enumerate(singleValueDefFields):
                tqData[tqId][singleValueDefFields[iP]] = record[iP]

        tqNeedCleaning = False
        for field in multiValueDefFields:
            table = "`tq_TQTo%s`" % field
            sqlCmd = "SELECT %s.TQId, %s.Value FROM %s" % (table, table, table)
            retVal = self._query(sqlCmd)
            if not retVal["OK"]:
                self.log.error("Can't retrieve task queues field", "%s info: %s" % (field, retVal["Message"]))
                return retVal
            for record in retVal["Value"]:
                tqId = record[0]
                value = record[1]
                if tqId not in tqData:
                    if tqIdList is None or tqId in tqIdList:
                        self.log.verbose(
                            "Task Queue is defined for a field, but does not exist: triggering a cleaning",
                            "TQID: %s, field: %s" % (tqId, field),
                        )
                        tqNeedCleaning = True
                else:
                    if field not in tqData[tqId]:
                        tqData[tqId][field] = []
                    tqData[tqId][field].append(value)
        if tqNeedCleaning:
            self.cleanOrphanedTaskQueues()
        return S_OK(tqData)

    def __updateGlobalShares(self):
        """
        Update internal structure for shares
        """
        # Update group shares
        self.__groupShares = self.getGroupShares()
        # Apply corrections if enabled
        if self.isSharesCorrectionEnabled():
            result = self.getGroupsInTQs()
            if not result["OK"]:
                self.log.error("Could not get groups in the TQs", result["Message"])
            activeGroups = result["Value"]
            newShares = {}
            for group in activeGroups:
                if group in self.__groupShares:
                    newShares[group] = self.__groupShares[group]
            newShares = self.__sharesCorrector.correctShares(newShares)
            for group in self.__groupShares:
                if group in newShares:
                    self.__groupShares[group] = newShares[group]

    def recalculateTQSharesForAll(self):
        """
        Recalculate all priorities for TQ's
        """
        if self.isSharesCorrectionEnabled():
            self.log.info("Updating correctors state")
            self.__sharesCorrector.update()
        self.__updateGlobalShares()
        self.log.info("Recalculating shares for all TQs")
        retVal = self._getConnection()
        if not retVal["OK"]:
            return S_ERROR("Can't insert job: %s" % retVal["Message"])
        result = self._query("SELECT DISTINCT( OwnerGroup ) FROM `tq_TaskQueues`")
        if not result["OK"]:
            return result
        for group in [r[0] for r in result["Value"]]:
            self.recalculateTQSharesForEntity("all", group)
        return S_OK()

    def recalculateTQSharesForEntity(self, userDN, userGroup, connObj=False):
        """
        Recalculate the shares for a userDN/userGroup combo
        """
        self.log.info("Recalculating shares", "for %s@%s TQs" % (userDN, userGroup))
        if userGroup in self.__groupShares:
            share = self.__groupShares[userGroup]
        else:
            share = float(DEFAULT_GROUP_SHARE)
        if Properties.JOB_SHARING in Registry.getPropertiesForGroup(userGroup):
            # If group has JobSharing just set prio for that entry, userDN is irrelevant
            return self.__setPrioritiesForEntity(userDN, userGroup, share, connObj=connObj)

        selSQL = "SELECT OwnerDN, COUNT(OwnerDN) FROM `tq_TaskQueues` WHERE OwnerGroup='%s' GROUP BY OwnerDN" % (
            userGroup
        )
        result = self._query(selSQL, conn=connObj)
        if not result["OK"]:
            return result
        # Get owners in this group and the amount of times they appear
        data = [(r[0], r[1]) for r in result["Value"] if r]
        numOwners = len(data)
        # If there are no owners do now
        if numOwners == 0:
            return S_OK()
        # Split the share amongst the number of owners
        share /= numOwners
        entitiesShares = dict([(row[0], share) for row in data])
        # If corrector is enabled let it work it's magic
        if self.isSharesCorrectionEnabled():
            entitiesShares = self.__sharesCorrector.correctShares(entitiesShares, group=userGroup)
        # Keep updating
        owners = dict(data)
        # IF the user is already known and has more than 1 tq, the rest of the users don't need to be modified
        # (The number of owners didn't change)
        if userDN in owners and owners[userDN] > 1:
            return self.__setPrioritiesForEntity(userDN, userGroup, entitiesShares[userDN], connObj=connObj)
        # Oops the number of owners may have changed so we recalculate the prio for all owners in the group
        for userDN in owners:
            self.__setPrioritiesForEntity(userDN, userGroup, entitiesShares[userDN], connObj=connObj)
        return S_OK()

    def __setPrioritiesForEntity(self, userDN, userGroup, share, connObj=False, consolidationFunc="AVG"):
        """
        Set the priority for a userDN/userGroup combo given a splitted share
        """
        self.log.info("Setting priorities", "to %s@%s TQs" % (userDN, userGroup))
        tqCond = ["t.OwnerGroup='%s'" % userGroup]
        allowBgTQs = gConfig.getValue("/Registry/Groups/%s/AllowBackgroundTQs" % userGroup, False)
        if Properties.JOB_SHARING not in Registry.getPropertiesForGroup(userGroup):
            res = self._escapeString(userDN)
            if not res["OK"]:
                return res
            userDN = res["Value"]
            tqCond.append("t.OwnerDN= %s " % userDN)
        tqCond.append("t.TQId = j.TQId")
        if consolidationFunc == "AVG":
            selectSQL = "SELECT j.TQId, SUM( j.RealPriority )/COUNT(j.RealPriority) \
FROM `tq_TaskQueues` t, `tq_Jobs` j WHERE "
        elif consolidationFunc == "SUM":
            selectSQL = "SELECT j.TQId, SUM( j.RealPriority ) FROM `tq_TaskQueues` t, `tq_Jobs` j WHERE "
        else:
            return S_ERROR("Unknown consolidation func %s for setting priorities" % consolidationFunc)
        selectSQL += " AND ".join(tqCond)
        selectSQL += " GROUP BY t.TQId"
        result = self._query(selectSQL, conn=connObj)
        if not result["OK"]:
            return result

        tqDict = dict(result["Value"])
        if not tqDict:
            return S_OK()
        # Calculate Sum of priorities
        totalPrio = 0
        for k in tqDict:
            if tqDict[k] > 0.1 or not allowBgTQs:
                totalPrio += tqDict[k]
        # Update prio for each TQ
        for tqId in tqDict:
            if tqDict[tqId] > 0.1 or not allowBgTQs:
                prio = (share / totalPrio) * tqDict[tqId]
            else:
                prio = TQ_MIN_SHARE
            prio = max(prio, TQ_MIN_SHARE)
            tqDict[tqId] = prio

        # Generate groups of TQs that will have the same prio=sum(prios) maomenos
        result = self.retrieveTaskQueues(list(tqDict))
        if not result["OK"]:
            return result
        allTQsData = result["Value"]
        tqGroups = {}
        for tqid in allTQsData:
            tqData = allTQsData[tqid]
            for field in ("Jobs", "Priority") + priorityIgnoredFields:
                if field in tqData:
                    tqData.pop(field)
            tqHash = []
            for f in sorted(tqData):
                tqHash.append("%s:%s" % (f, tqData[f]))
            tqHash = "|".join(tqHash)
            if tqHash not in tqGroups:
                tqGroups[tqHash] = []
            tqGroups[tqHash].append(tqid)
        tqGroups = [tqGroups[td] for td in tqGroups]

        # Do the grouping
        for tqGroup in tqGroups:
            totalPrio = 0
            if len(tqGroup) < 2:
                continue
            for tqid in tqGroup:
                totalPrio += tqDict[tqid]
            for tqid in tqGroup:
                tqDict[tqid] = totalPrio

        # Group by priorities
        prioDict = {}
        for tqId in tqDict:
            prio = tqDict[tqId]
            if prio not in prioDict:
                prioDict[prio] = []
            prioDict[prio].append(tqId)

        # Execute updates
        for prio in prioDict:
            tqList = ", ".join([str(tqId) for tqId in prioDict[prio]])
            updateSQL = "UPDATE `tq_TaskQueues` SET Priority=%.4f WHERE TQId in ( %s )" % (prio, tqList)
            self._update(updateSQL, conn=connObj)
        return S_OK()

    @staticmethod
    def getGroupShares():
        """
        Get all the shares as a DICT
        """
        result = gConfig.getSections("/Registry/Groups")
        if result["OK"]:
            groups = result["Value"]
        else:
            groups = []
        shares = {}
        for group in groups:
            shares[group] = gConfig.getValue("/Registry/Groups/%s/JobShare" % group, DEFAULT_GROUP_SHARE)
        return shares



"""HTTPS module based on the GFAL2_StorageBase class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# from DIRAC
from DIRAC.Resources.Storage.GFAL2_StorageBase import GFAL2_StorageBase
from DIRAC import gLogger


class GFAL2_HTTPSStorage(GFAL2_StorageBase):

    """.. class:: GFAL2_HTTPSStorage

    HTTP interface to StorageElement using gfal2
    """

    # davs is for https with direct access + third party
    _INPUT_PROTOCOLS = ["file", "http", "https", "dav", "davs"]
    _OUTPUT_PROTOCOLS = ["http", "https", "dav", "davs"]

    def __init__(self, storageName, parameters):
        """c'tor"""
        # # init base class
        super(GFAL2_HTTPSStorage, self).__init__(storageName, parameters)
        self.srmSpecificParse = False

        self.log = gLogger.getSubLogger("GFAL2_HTTPSStorage")

        self.pluginName = "GFAL2_HTTPS"

        # We don't need extended attributes for metadata
        self._defaultExtendedAttributes = None



""" The RucioFileCatalogClient is a class that allows to interface Dirac with Rucio
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id"

import os
import uuid
import datetime
from copy import deepcopy

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase

from rucio.client import Client
from rucio.common.exception import (
    DataIdentifierNotFound,
    FileReplicaAlreadyExists,
    CannotAuthenticate,
    MissingClientParameter,
)
from rucio.common.utils import chunks, extract_scope

sLog = gLogger.getSubLogger(__name__)


def get_scope(lfn, scopes=None):
    """Helper function that extract the scope from the LFN"""
    if scopes is None:
        scopes = []
    scope, _ = extract_scope(did=lfn, scopes=scopes)
    return scope


class RucioFileCatalogClient(FileCatalogClientBase):

    READ_METHODS = FileCatalogClientBase.READ_METHODS + [
        "isLink",
        "readLink",
        "isFile",
        "getFileMetadata",
        "getReplicas",
        "getReplicaStatus",
        "getFileSize",
        "isDirectory",
        "getDirectoryReplicas",
        "listDirectory",
        "getDirectoryMetadata",
        "getDirectorySize",
        "getDirectoryContents",
        "resolveDataset",
        "getLFNForPFN",
        "getUserDirectory",
    ]

    WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + [
        "createLink",
        "removeLink",
        "addFile",
        "addReplica",
        "removeReplica",
        "removeFile",
        "setReplicaStatus",
        "setReplicaHost",
        "createDirectory",
        "removeDirectory",
        "removeDataset",
        "removeFileFromDataset",
        "createDataset",
        "changePathOwner",
        "changePathMode",
    ]

    NO_LFN_METHODS = FileCatalogClientBase.NO_LFN_METHODS + [
        "getUserDirectory",
        "createUserDirectory",
        "createUserMapping",
        "removeUserDirectory",
    ]

    ADMIN_METHODS = FileCatalogClientBase.ADMIN_METHODS + [
        "getUserDirectory",
        "createUserDirectory",
        "createUserMapping",
        "removeUserDirectory",
    ]

    def __init__(self, **options):
        """Constructor"""
        self.convertUnicode = True
        proxyInfo = {"OK": False}
        if os.getenv("RUCIO_AUTH_TYPE") == "x509_proxy" and not os.getenv("X509_USER_PROXY"):
            proxyInfo = getProxyInfo(disableVOMS=True)
            if proxyInfo["OK"]:
                os.environ["X509_USER_PROXY"] = proxyInfo["Value"]["path"]
                sLog.debug("X509_USER_PROXY not defined. Using %s" % proxyInfo["Value"]["path"])
        try:
            self._client = Client()
            self.account = self._client.account
        except (CannotAuthenticate, MissingClientParameter):
            if os.getenv("RUCIO_AUTH_TYPE") == "x509_proxy":
                if not proxyInfo["OK"]:
                    proxyInfo = getProxyInfo(disableVOMS=True)
                if proxyInfo["OK"]:
                    dn = proxyInfo["Value"]["identity"]
                    username = proxyInfo["Value"]["username"]
                    self.account = username
                    sLog.debug("Switching to account %s mapped to proxy %s" % (username, dn))

        try:
            self._client = Client(account=self.account)
            self.scopes = self._client.list_scopes()
        except Exception as err:
            sLog.error("Cannot instantiate RucioFileCatalog interface", "error : %s" % repr(err))

    @property
    def client(self):
        """Check if the session to the server is still active and return an instance of RucioClient"""
        try:
            self._client.ping()
            return self._client
        except Exception:
            self._client = Client(account=self.account)
            self.scopes = self._client.list_scopes()
            return self._client

    def __getDidsFromLfn(self, lfn):
        """Helper function to convert LFNs into DIDs"""
        if lfn.find(":") > -1:
            scope, name = lfn.split(":")
        else:
            scope = get_scope(lfn, scopes=self.scopes)
            name = lfn
        return {"scope": scope, "name": name}

    @checkCatalogArguments
    def getCompatibleMetadata(self, queryDict, path, credDict):
        """Get distinct metadata values compatible with the given already defined metadata"""
        if path != "/":
            result = self.exists(path)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Path not found: %s" % path)
        result = S_OK({})
        return result

    @checkCatalogArguments
    def getReplicas(self, lfns, allStatus=False):
        """Returns replicas for an LFN or list of LFNs"""
        result = {"OK": True, "Value": {"Successful": {}, "Failed": {}}}
        lfnChunks = breakListIntoChunks(lfns, 1000)

        for lfnList in lfnChunks:
            try:
                didList = [self.__getDidsFromLfn(lfn) for lfn in lfnList if lfn]
                for rep in self.client.list_replicas(didList):
                    if rep:
                        lfn = rep["name"]
                        if self.convertUnicode:
                            lfn = str(lfn)
                        if lfn not in result["Value"]["Successful"]:
                            result["Value"]["Successful"][lfn] = {}
                        for rse in rep["rses"]:
                            if self.convertUnicode:
                                result["Value"]["Successful"][lfn][str(rse)] = str(rep["rses"][rse][0])
                            else:
                                result["Value"]["Successful"][lfn][rse] = rep["rses"][rse][0]
                    else:
                        for did in didList:
                            result["Value"]["Failed"][did["name"]] = "Error"
            except Exception as err:
                return S_ERROR(str(err))
        return result

    @checkCatalogArguments
    def listDirectory(self, lfns, verbose=False):
        """Returns the result of __getDirectoryContents for multiple supplied paths"""
        result = {"OK": True, "Value": {"Successful": {}, "Failed": {}}}

        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                # First need to check if it's a dataset or container
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "CONTAINER":
                    if lfn not in result["Value"]["Successful"]:
                        result["Value"]["Successful"][lfn] = {"Files": {}, "Links": {}, "SubDirs": {}}
                    for child in self.client.list_content(scope=did["scope"], name=did["name"]):
                        childName = child["name"]
                        if self.convertUnicode:
                            childName = str(childName)
                        if child["type"] in ["DATASET", "CONTAINER"]:
                            result["Value"]["Successful"][lfn]["SubDirs"][childName] = {"Mode": 509}
                            if verbose:
                                result["Value"]["Successful"][lfn]["SubDirs"][childName] = {
                                    "Status": "-",
                                    "CreationDate": datetime.datetime(1970, 1, 1),
                                    "ChecksumType": "",
                                    "OwnerRole": "None",
                                    "Checksum": "",
                                    "NumberOfLinks": -999,
                                    "OwnerDN": "None",
                                    "Mode": 509,
                                    "ModificationDate": datetime.datetime(1970, 1, 1),
                                    "GUID": "",
                                    "LastAccess": 0,
                                    "Size": 0,
                                }
                        else:
                            result["Value"]["Successful"][lfn]["Files"][childName] = {"Mode": 509}
                            if verbose:
                                pass
                elif meta["did_type"] == "DATASET":
                    file_dict = {}
                    for file_did in self.client.list_files(scope=did["scope"], name=did["name"]):
                        guid = file_did["guid"]
                        file_dict[file_did["name"]] = guid
                        if guid:
                            file_dict[file_did["name"]] = str(uuid.UUID(guid))
                    if lfn not in result["Value"]["Successful"]:
                        result["Value"]["Successful"][lfn] = {"Files": {}, "Links": {}, "SubDirs": {}}
                    for rep in self.client.list_replicas(
                        [
                            did,
                        ]
                    ):
                        if rep:
                            name = rep["name"]
                            if self.convertUnicode:
                                name = str(name)
                            result["Value"]["Successful"][lfn]["Files"][name] = {"Metadata": {}, "Replicas": {}}
                            result["Value"]["Successful"][lfn]["Files"][name]["Metadata"] = {
                                "GUID": str(file_dict.get(name, "UNKNOWN")),
                                "Mode": 436,
                                "Size": rep["bytes"],
                            }
                            if verbose:
                                result["Value"]["Successful"][lfn]["Files"][name]["Metadata"] = {
                                    "Status": "-",
                                    "CreationDate": datetime.datetime(1970, 1, 1),
                                    "ChecksumType": "AD",
                                    "OwnerRole": "None",
                                    "Checksum": str(rep["adler32"]),
                                    "NumberOfLinks": 1,
                                    "OwnerDN": "None",
                                    "Mode": 436,
                                    "ModificationDate": datetime.datetime(1970, 1, 1),
                                    "GUID": str(file_dict.get(name, "UNKNOWN")),
                                    "LastAccess": 0,
                                    "Size": rep["bytes"],
                                }
                            for rse in rep["rses"]:
                                pfn = rep["rses"][rse][0]
                                if self.convertUnicode:
                                    pfn = str(pfn)
                                result["Value"]["Successful"][lfn]["Files"][name]["Replicas"][rse] = {
                                    "PFN": pfn,
                                    "Status": "U",
                                }
            except DataIdentifierNotFound as err:
                result["Value"]["Failed"][lfn] = "No such file or directory"
            except Exception as err:
                return S_ERROR(str(err))
        return result

    @checkCatalogArguments
    def getFileMetadata(self, lfns, ownership=False):
        """Returns the file metadata associated to a supplied LFN"""
        successful, failed = {}, {}
        lfnChunks = breakListIntoChunks(lfns, 1000)
        listFiles = deepcopy(list(lfns))
        for chunk in lfnChunks:
            try:
                dids = [self.__getDidsFromLfn(lfn) for lfn in chunk]
                for meta in self.client.get_metadata_bulk(dids):
                    lfn = str(meta["name"])
                    if meta["did_type"] in ["DATASET", "CONTAINER"]:
                        nlinks = len([child for child in self.client.list_content(meta["scope"], meta["name"])])
                        successful[lfn] = {
                            "Checksum": "",
                            "ChecksumType": "",
                            "CreationDate": meta["created_at"],
                            "GUID": "",
                            "Mode": 509,
                            "ModificationDate": meta["updated_at"],
                            "NumberOfLinks": nlinks,
                            "Size": 0,
                            "Status": "-",
                        }
                        try:
                            listFiles.remove(lfn)
                        except ValueError:
                            pass
                    else:
                        guid = meta["guid"]
                        if guid:
                            guid = str(uuid.UUID(guid))
                        successful[lfn] = {
                            "Checksum": str(meta["adler32"]),
                            "ChecksumType": "AD",
                            "CreationDate": meta["created_at"],
                            "GUID": guid,
                            "Mode": 436,
                            "ModificationDate": meta["updated_at"],
                            "NumberOfLinks": 1,
                            "Size": meta["bytes"],
                            "Status": "-",
                        }
                        try:
                            listFiles.remove(lfn)
                        except ValueError:
                            pass
            except DataIdentifierNotFound as err:
                failed[lfn] = str(err)
            except Exception as err:
                return S_ERROR(str(err))
        for lfn in listFiles:
            failed[lfn] = "No such file or directory"
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def exists(self, lfns):
        """Check if the path exists"""
        result = {"OK": True, "Value": {"Successful": {}, "Failed": {}}}
        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                exists = True
                self.client.get_metadata(did["scope"], did["name"])
            except DataIdentifierNotFound:
                exists = False
            except Exception as err:
                return S_ERROR(str(err))
            result["Value"]["Successful"][lfn] = exists
        return result

    @checkCatalogArguments
    def getFileSize(self, lfns):
        """Get the size of a supplied file"""
        result = {"OK": True, "Value": {"Successful": {}, "Failed": {}}}
        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "FILE":
                    result["Value"]["Successful"][lfn] = meta["bytes"]
                else:
                    result["Value"]["Successful"][lfn] = 0
            except DataIdentifierNotFound:
                result["Value"]["Failed"][lfn] = "No such file or directory"
            except Exception as err:
                return S_ERROR(str(err))
        return result

    @checkCatalogArguments
    def isDirectory(self, lfns):
        """Determine whether the path is a directory"""
        result = {"Successful": {}, "Failed": {}}
        dids = [self.__getDidsFromLfn(lfn) for lfn in lfns]
        try:
            for meta in self.client.get_metadata_bulk(dids):
                lfn = str(meta["name"])
                result["Successful"][lfn] = meta["did_type"] in ["DATASET", "CONTAINER"]
            for lfn in lfns:
                if lfn not in result["Successful"] and lfn not in result["Failed"]:
                    result["Failed"][lfn] = "No such file or directory"
        except Exception as err:
            return S_ERROR(str(err))
        return S_OK(result)

    @checkCatalogArguments
    def isFile(self, lfns):
        """Determine whether the path is a file"""
        result = {"Successful": {}, "Failed": {}}
        dids = []
        for lfn in lfns:
            dids.append(self.__getDidsFromLfn(lfn))
        try:
            for meta in self.client.get_metadata_bulk(dids):
                lfn = str(meta["name"])
                result["Successful"][lfn] = meta["did_type"] in ["FILE"]
            for lfn in lfns:
                if lfn not in result["Successful"] and lfn not in result["Failed"]:
                    result["Failed"][lfn] = "No such file or directory"
        except Exception as err:
            return S_ERROR(str(err))
        return S_OK(result)

    @checkCatalogArguments
    def addFile(self, lfns):
        """Register supplied files"""
        failed = {}
        successful = {}
        deterministicDictionary = {}
        for lfnList in breakListIntoChunks(lfns, 100):
            listLFNs = []
            for lfn in list(lfnList):
                lfnInfo = lfns[lfn]
                pfn = None
                se = lfnInfo["SE"]
                if se not in deterministicDictionary:
                    isDeterministic = self.client.get_rse(se)["deterministic"]
                    deterministicDictionary[se] = isDeterministic
                if not deterministicDictionary[se]:
                    pfn = lfnInfo["PFN"]
                size = lfnInfo["Size"]
                guid = lfnInfo.get("GUID", None)
                checksum = lfnInfo["Checksum"]
                rep = {"lfn": lfn, "bytes": size, "adler32": checksum, "rse": se}
                if pfn:
                    rep["pfn"] = pfn
                if guid:
                    rep["guid"] = guid
                listLFNs.append(rep)
            try:
                self.client.add_files(lfns=listLFNs, ignore_availability=True)
                for lfn in list(lfnList):
                    successful[lfn] = True
            except Exception as err:
                # Try inserting one by one
                sLog.warning("Cannot bulk insert files", "error : %s" % repr(err))
                for lfn in list(lfnList):
                    try:
                        self.client.add_files(lfns=lfn, ignore_availability=True)
                        successful[lfn] = True
                    except FileReplicaAlreadyExists:
                        successful[lfn] = True
                    except Exception as err:
                        failed[lfn] = str(err)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def addReplica(self, lfns):
        """This adds a replica to the catalogue."""
        failed = {}
        successful = {}
        deterministicDictionary = {}
        for lfn, info in lfns.items():
            pfn = None
            se = info["SE"]
            if se not in deterministicDictionary:
                isDeterministic = self.client.get_rse(se)["deterministic"]
                deterministicDictionary[se] = isDeterministic
            if not deterministicDictionary[se]:
                pfn = info["PFN"]
            size = info.get("Size", None)
            checksum = info.get("Checksum", None)
            try:
                did = self.__getDidsFromLfn(lfn)
                if not size or not checksum:
                    meta = self.client.get_metadata(did["scope"], did["name"])
                    size = meta["bytes"]
                    checksum = meta["adler32"]
                rep = {"scope": did["scope"], "name": did["name"], "bytes": size, "adler32": checksum}
                if pfn:
                    rep["pfn"] = pfn
                self.client.add_replicas(
                    rse=se,
                    files=[
                        rep,
                    ],
                )
                self.client.add_replication_rule(
                    [{"scope": did["scope"], "name": did["name"]}],
                    copies=1,
                    activity="User Transfers",
                    rse_expression=se,
                    weight=None,
                    lifetime=None,
                    grouping="NONE",
                    account=self.account,
                )
                successful[lfn] = True
            except FileReplicaAlreadyExists:
                successful[lfn] = True
            except Exception as err:
                failed[lfn] = str(err)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeReplica(self, lfns):
        """Remove the supplied replicas"""
        failed = {}
        successful = {}
        for lfn, info in lfns.items():
            if "SE" not in info:
                failed[lfn] = "Required parameters not supplied"
                continue
            se = info["SE"]
            try:
                did = self.__getDidsFromLfn(lfn)
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "FILE":
                    # For file cannot use dataset_locks to identify the rule
                    for rule in self.client.list_did_rules(did["scope"], did["name"]):
                        rid = rule["id"]
                        self.client.update_replication_rule(rid, options={"lifetime": -86400})
                    successful[lfn] = True
                elif meta["did_type"] == "DATASET":
                    rules = {}
                    for lock in self.client.get_dataset_locks(did["scope"], did["name"]):
                        rule_id = lock["rule_id"]
                        if rule_id not in rules:
                            rules[rule_id] = []
                        rse = lock["rse"]
                        if str(rse) not in rules[rule_id]:
                            rules[rule_id].append(str(rse))

                    for rid in rules:
                        if len(rules[rid]) == 1 and rules[rid][0] == se:
                            rule_info = self.client.get_replication_rule(rid)
                            if str(rule_info["name"]) == lfn:
                                self.client.update_replication_rule(rid, options={"lifetime": -86400})
                    successful[lfn] = True

            except Exception as err:
                # Always assumes that it succeeded
                failed[lfn] = str(err)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeFile(self, lfns):
        """Remove the supplied path"""
        resDict = {"Successful": {}, "Failed": {}}
        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "FILE":
                    parentLfn = "/".join(lfn.split("/")[:-1])
                    parentDid = self.__getDidsFromLfn(parentLfn)
                    dsnScope, dsnName = parentDid["scope"], parentDid["name"]
                    try:
                        self.client.detach_dids(dsnScope, dsnName, [{"scope": did["scope"], "name": did["name"]}])
                        resDict["Successful"][lfn] = True
                    except Exception as err:
                        resDict["Failed"][lfn] = str(err)
                else:
                    resDict["Failed"][lfn] = "Not a file"
            except DataIdentifierNotFound:
                resDict["Failed"][lfn] = "No such file or directory"
            except Exception as err:
                return S_ERROR(str(err))
        return S_OK(resDict)

    @checkCatalogArguments
    def removeDirectory(self, lfns):
        """Remove the supplied directory"""
        resDict = {"Successful": {}, "Failed": {}}
        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "DATASET":
                    try:
                        self.client.set_metadata(scope=did["scope"], name=did["name"], key="lifetime", value=1)
                        resDict["Successful"][lfn] = True
                    except Exception as err:
                        resDict["Failed"][lfn] = str(err)
                else:
                    resDict["Failed"][lfn] = "Not a directory"
            except DataIdentifierNotFound:
                resDict["Failed"][lfn] = "No such file or directory"
            except Exception as err:
                return S_ERROR(str(err))
        return S_OK(resDict)

    @checkCatalogArguments
    def getDirectorySize(self, lfns, longOutput=False, rawFiles=False):
        """Get the directory size"""
        resDict = {"Successful": {}, "Failed": {}}
        for lfn in lfns:
            try:
                did = self.__getDidsFromLfn(lfn)
                meta = self.client.get_metadata(did["scope"], did["name"])
                if meta["did_type"] == "FILE":
                    resDict["Failed"][lfn] = "Not a directory"

                elif meta["did_type"] == "CONTAINER":
                    resDict["Successful"][lfn] = {
                        "ClosedDirs": [],
                        "Files": 0,
                        "SiteUsage": {},
                        "SubDirs": {},
                        "TotalSize": 0,
                    }
                    for child in self.client.list_content(scope=did["scope"], name=did["name"]):
                        childName = child["name"]
                        if self.convertUnicode:
                            childName = str(childName)
                        resDict["Successful"][lfn]["SubDirs"][childName] = datetime.datetime.now()

                else:
                    resDict["Successful"][lfn] = {
                        "ClosedDirs": [],
                        "Files": 0,
                        "SiteUsage": {},
                        "SubDirs": {},
                        "TotalSize": 0,
                    }
                    for child in self.client.list_files(scope=did["scope"], name=did["name"]):
                        resDict["Successful"][lfn]["Files"] += 1
                        resDict["Successful"][lfn]["TotalSize"] += child["bytes"]
                    for rep in self.client.list_replicas([{"scope": did["scope"], "name": did["name"]}]):
                        fileSize = rep["bytes"]
                        for rse in rep["rses"]:
                            if rse not in resDict["Successful"][lfn]["SiteUsage"]:
                                resDict["Successful"][lfn]["SiteUsage"][rse] = {"Files": 0, "Size": 0}
                            resDict["Successful"][lfn]["SiteUsage"][rse]["Files"] += 1
                            resDict["Successful"][lfn]["SiteUsage"][rse]["Size"] += fileSize

            except DataIdentifierNotFound:
                resDict["Failed"][lfn] = "No such file or directory"
            except Exception as err:
                return S_ERROR(str(err))
        return S_OK(resDict)



#!/usr/bin/env python
"""
Script that helps setting the token of the elements in RSS.
It can acquire or release the token.

If the releaseToken switch is used, no matter what was the previous token, it will be set to rs_svc (RSS owns it).
If not set, the token will be set to whatever username is defined on the proxy loaded while issuing
this command. In the second case, the token lasts one day.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

# DIRAC
from DIRAC import gLogger, exit as DIRACExit, S_OK, version
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient


subLogger = None
switchDict = {}


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    switches = (
        ("element=", "Element family to be Synchronized ( Site, Resource or Node )"),
        ("name=", "Name, name of the element where the change applies"),
        ("statusType=", "StatusType, if none applies to all possible statusTypes"),
        ("reason=", "Reason to set the Status"),
        ("days=", "Number of days the token is acquired"),
        ("releaseToken", "Release the token and let the RSS take control"),
        ("VO=", "VO to set a token for. When omitted token will be set for all VOs"),
    )

    for switch in switches:
        Script.registerSwitch("", switch[0], switch[1])


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """

    usageMessage = "  DIRAC %s\n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if args:
        subLogger.error("Found the following positional args '%s', but we only accept switches" % args)
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    switches = dict(Script.getUnprocessedSwitches())
    switches.setdefault("statusType", None)
    switches.setdefault("VO", None)
    switches.setdefault("days", 1)
    if "releaseToken" in switches:
        switches["releaseToken"] = True
    else:
        switches["releaseToken"] = False

    for key in ("element", "name", "reason"):

        if key not in switches:
            subLogger.error("%s Switch missing" % key)
            subLogger.error("Please, check documentation above")
            Script.showHelp(exitCode=1)

    if not switches["element"] in ("Site", "Resource", "Node"):
        subLogger.error("Found %s as element switch" % switches["element"])
        subLogger.error("Please, check documentation above")
        Script.showHelp(exitCode=1)

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return switches


def proxyUser():
    """
    Read proxy to get username.
    """

    res = getProxyInfo()
    if not res["OK"]:
        return res

    return S_OK(res["Value"]["username"])


def setToken(user):
    """
    Function that gets the user token, sets the validity for it. Gets the elements
    in the database for a given name and statusType(s). Then updates the status
    of all them adding a reason and the token.
    """

    rssClient = ResourceStatusClient()

    # This is a little bit of a nonsense, and certainly needs to be improved.
    # To modify a list of elements, we have to do it one by one. However, the
    # modify method does not discover the StatusTypes ( which in this script is
    # an optional parameter ). So, we get them from the DB and iterate over them.
    elements = rssClient.selectStatusElement(
        switchDict["element"],
        "Status",
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        vO=switchDict["VO"],
        meta={"columns": ["StatusType", "TokenOwner"]},
    )

    if not elements["OK"]:
        return elements
    elements = elements["Value"]

    # If there list is empty they do not exist on the DB !
    if not elements:
        subLogger.warn(
            "Nothing found for %s, %s, %s %s"
            % (switchDict["element"], switchDict["name"], switchDict["VO"], switchDict["statusType"])
        )
        return S_OK()

    # If we want to release the token
    if switchDict["releaseToken"]:
        tokenExpiration = datetime.max
        newTokenOwner = "rs_svc"
    else:
        tokenExpiration = datetime.utcnow().replace(microsecond=0) + timedelta(days=int(switchDict["days"]))
        newTokenOwner = user

    subLogger.always("New token: %s --- until %s" % (newTokenOwner, tokenExpiration))

    for statusType, tokenOwner in elements:

        # If a user different than the one issuing the command and RSS
        if tokenOwner != user and tokenOwner != "rs_svc":
            subLogger.info("%s(%s) belongs to the user: %s" % (switchDict["name"], statusType, tokenOwner))

        # does the job
        result = rssClient.modifyStatusElement(
            switchDict["element"],
            "Status",
            name=switchDict["name"],
            statusType=statusType,
            reason=switchDict["reason"],
            tokenOwner=newTokenOwner,
            vO=switchDict["VO"],
            tokenExpiration=tokenExpiration,
        )
        if not result["OK"]:
            return result

        if tokenOwner == newTokenOwner:
            msg = "(extended)"
        elif newTokenOwner == "rs_svc":
            msg = "(released)"
        else:
            msg = "(aquired from %s)" % tokenOwner

        subLogger.info("name:%s, VO:%s statusType:%s %s" % (switchDict["name"], switchDict["VO"], statusType, msg))
    return S_OK()


@Script()
def main():
    """
    Main function of the script. Gets the username from the proxy loaded and sets
    the token taking into account that user and the switchDict parameters.
    """
    global subLogger
    global switchDict

    # Logger initialization
    subLogger = gLogger.getSubLogger(__file__)

    # Script initialization
    registerSwitches()
    registerUsageMessage()
    switchDict = parseSwitches()

    user = proxyUser()
    if not user["OK"]:
        subLogger.error(user["Message"])
        DIRACExit(1)
    user = user["Value"]

    res = setToken(user)
    if not res["OK"]:
        subLogger.error(res["Message"])
        DIRACExit(1)

    DIRACExit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac_setup_site.py
# Author :  Ricardo Graciani
########################################################################
"""
Initial installation and configuration of a new DIRAC server (DBs, Services, Agents, Web Portal,...)
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


class Params(object):
    def __init__(self):
        self.exitOnError = False

    def setExitOnError(self, value):
        self.exitOnError = True
        return S_OK()


@Script()
def main():
    cliParams = Params()

    Script.disableCS()
    Script.registerSwitch(
        "e", "exitOnError", "flag to exit on error of any component installation", cliParams.setExitOnError
    )

    Script.addDefaultOptionValue("/DIRAC/Security/UseServerCertificate", "yes")
    Script.addDefaultOptionValue("LogLevel", "INFO")
    Script.parseCommandLine()
    args = Script.getExtraCLICFGFiles()

    if len(args) > 1:
        Script.showHelp(exitCode=1)

    cfg = None
    if len(args):
        cfg = args[0]
    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    gComponentInstaller.exitOnError = cliParams.exitOnError

    result = gComponentInstaller.setupSite(Script.localCfg, cfg)
    if not result["OK"]:
        print("ERROR:", result["Message"])
        exit(-1)

    result = gComponentInstaller.getStartupComponentStatus([])
    if not result["OK"]:
        print("ERROR:", result["Message"])
        exit(-1)

    print("\nStatus of installed components:\n")
    result = gComponentInstaller.printStartupStatus(result["Value"])
    if not result["OK"]:
        print("ERROR:", result["Message"])
        exit(-1)


if __name__ == "__main__":
    main()



########################################################################
# $HeadURL $
# File: RequestValidatorTests.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2012/09/25 13:49:20
########################################################################

""" :mod: RequestValidatorTests
    =======================

    .. module: RequestValidatorTests
    :synopsis: test cases for RequestValidator
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    test cases for RequestValidator
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

##
# @file RequestValidatorTests.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/09/25 13:49:31
# @brief Definition of RequestValidatorTests class.

## imports
import unittest

## from DIRAC
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

## SUT
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator


########################################################################
class RequestValidatorTests(unittest.TestCase):
    """
    .. class:: RequestValidatorTests

    """

    def setUp(self):
        """test setup"""
        self.request = Request()
        self.operation = Operation()
        self.file = File()

    def tearDown(self):
        """test tear down"""
        del self.request
        del self.operation
        del self.file

    def testValidator(self):
        """validator test"""

        ## create validator
        validator = RequestValidator()
        self.assertEqual(isinstance(validator, RequestValidator), True)

        ## RequestName not set
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"Message": "RequestName not set", "OK": False})
        self.request.RequestName = "test_request"

        # # no operations
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"Message": "Operations not present in request 'test_request'", "OK": False})
        self.request.addOperation(self.operation)

        # # type not set
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"Message": "Operation #0 in request 'test_request' hasn't got Type set", "OK": False})
        self.operation.Type = "ReplicateAndRegister"

        # # files not present
        ret = validator.validate(self.request)
        self.assertEqual(
            ret, {"Message": "Operation #0 of type 'ReplicateAndRegister' hasn't got files to process.", "OK": False}
        )
        self.operation.addFile(self.file)

        # # targetSE not set
        ret = validator.validate(self.request)
        self.assertEqual(
            ret, {"Message": "Operation #0 of type 'ReplicateAndRegister' is missing TargetSE attribute.", "OK": False}
        )
        self.operation.TargetSE = "CERN-USER"

        # # missing LFN
        ret = validator.validate(self.request)
        self.assertEqual(
            ret,
            {"Message": "Operation #0 of type 'ReplicateAndRegister' is missing LFN attribute for file.", "OK": False},
        )
        self.file.LFN = "/a/b/c"

        # # no ownerDN
        # force no owner DN because it takes the one of the current user
        self.request.OwnerDN = ""
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"Message": "Request 'test_request' is missing OwnerDN value", "OK": False})
        self.request.OwnerDN = "foo/bar=baz"

        # # no owner group
        # same, force it
        self.request.OwnerGroup = ""
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"Message": "Request 'test_request' is missing OwnerGroup value", "OK": False})
        self.request.OwnerGroup = "dirac_user"

        ## Checksum set, ChecksumType not set
        self.file.Checksum = "abcdef"
        ret = validator.validate(self.request)
        self.assertEqual(
            ret, {"Message": "File in operation #0 is missing Checksum (abcdef) or ChecksumType ()", "OK": False}
        )

        ## ChecksumType set, Checksum not set
        self.file.Checksum = ""
        self.file.ChecksumType = "adler32"

        ret = validator.validate(self.request)
        self.assertEqual(
            ret, {"Message": "File in operation #0 is missing Checksum () or ChecksumType (ADLER32)", "OK": False}
        )

        ## both set
        self.file.Checksum = "abcdef"
        self.file.ChecksumType = "adler32"
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"OK": True, "Value": None})

        ## both unset
        self.file.Checksum = ""
        self.file.ChecksumType = None
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"OK": True, "Value": None})

        ## all OK
        ret = validator.validate(self.request)
        self.assertEqual(ret, {"OK": True, "Value": None})


## test suite execution
if __name__ == "__main__":
    gTestLoader = unittest.TestLoader()
    gSuite = gTestLoader.loadTestsFromTestCase(RequestValidatorTests)
    gSuite = unittest.TestSuite([gSuite])
    unittest.TextTestRunner(verbosity=3).run(gSuite)



#!/bin/env python
"""
tests for HTCondorCEComputingElement module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest

from DIRAC.Resources.Computing import HTCondorCEComputingElement as HTCE
from DIRAC.Resources.Computing.BatchSystems import Condor
from DIRAC import S_OK

MODNAME = "DIRAC.Resources.Computing.HTCondorCEComputingElement"

STATUS_LINES = """
123.2 5
123.1 3
""".strip().split(
    "\n"
)

HISTORY_LINES = """
123 0 4
""".strip().split(
    "\n"
)


@pytest.fixture
def setUp():
    return {"Queue": "espresso", "GridEnv": "/dev/null"}


def test_parseCondorStatus():
    statusLines = """
  104097.9 2
  104098.0 1
  104098.1 4

  foo bar
  104098.2 3
  104098.3 5
  104098.4 7
  """.strip().split(
        "\n"
    )
    # force there to be an empty line

    expectedResults = {
        "104097.9": "Running",
        "104098.0": "Waiting",
        "104098.1": "Done",
        "104098.2": "Aborted",
        "104098.3": "HELD",
        "104098.4": "Unknown",
    }

    for jobID, expected in expectedResults.items():
        assert HTCE.parseCondorStatus(statusLines, jobID) == expected


def test_getJobStatus(mocker):
    """Test HTCondorCE getJobStatus"""
    mocker.patch(MODNAME + ".commands.getstatusoutput", side_effect=([(0, "\n".join(STATUS_LINES)), (0, 0)]))
    patchPopen = mocker.patch("DIRAC.Resources.Computing.BatchSystems.Condor.subprocess.Popen")
    patchPopen.return_value.communicate.side_effect = [("\n".join(HISTORY_LINES), "")]
    patchPopen.return_value.returncode = 0
    mocker.patch(MODNAME + ".HTCondorCEComputingElement._HTCondorCEComputingElement__cleanup")

    htce = HTCE.HTCondorCEComputingElement(12345)
    ret = htce.getJobStatus(
        [
            "htcondorce://condorce.foo.arg/123.0:::abc321",
            "htcondorce://condorce.foo.arg/123.1:::c3b2a1",
            "htcondorce://condorce.foo.arg/123.2:::c3b2a2",
            "htcondorce://condorce.foo.arg/333.3:::c3b2a3",
        ]
    )

    expectedResults = {
        "htcondorce://condorce.foo.arg/123.0": "Done",
        "htcondorce://condorce.foo.arg/123.1": "Aborted",
        "htcondorce://condorce.foo.arg/123.2": "Aborted",
        "htcondorce://condorce.foo.arg/333.3": "Unknown",
    }

    assert ret["OK"] is True
    assert expectedResults == ret["Value"]


def test_getJobStatusBatchSystem(mocker):
    """Test Condor Batch System plugin getJobStatus"""
    patchPopen = mocker.patch("DIRAC.Resources.Computing.BatchSystems.Condor.subprocess.Popen")
    patchPopen.return_value.communicate.side_effect = [("\n".join(STATUS_LINES), ""), ("\n".join(HISTORY_LINES), "")]
    patchPopen.return_value.returncode = 0

    with patchPopen:
        ret = Condor.Condor().getJobStatus(JobIDList=["123.0", "123.1", "123.2", "333.3"])

    expectedResults = {
        "123.0": "Done",
        "123.1": "Aborted",
        "123.2": "Unknown",  # HELD is treated as Unknown
        "333.3": "Unknown",
    }

    assert ret["Status"] == 0
    assert expectedResults == ret["Jobs"]


@pytest.mark.parametrize(
    "localSchedd, optionsNotExpected, optionsExpected",
    [
        (False, ["ShouldTransferFiles = YES", "WhenToTransferOutput = ON_EXIT_OR_EVICT"], ["universe = vanilla"]),
        (True, [], ["ShouldTransferFiles = YES", "WhenToTransferOutput = ON_EXIT_OR_EVICT", "universe = grid"]),
    ],
)
def test__writeSub(mocker, localSchedd, optionsNotExpected, optionsExpected):
    htce = HTCE.HTCondorCEComputingElement(12345)
    htce.useLocalSchedd = localSchedd
    subFileMock = mocker.Mock()

    patchFdopen = mocker.patch(MODNAME + ".os.fdopen", return_value=subFileMock)
    patchMkstemp = mocker.patch(MODNAME + ".tempfile.mkstemp", return_value=("os", "pilotName"))
    patchMkdir = mocker.patch(MODNAME + ".mkDir")

    with patchFdopen, patchMkstemp, patchMkdir:
        htce._HTCondorCEComputingElement__writeSub("dirac-install", 42, "", 1)  # pylint: disable=E1101
        for option in optionsNotExpected:
            # the three [0] are: call_args_list[firstCall][ArgsArgumentsTuple][FirstArgsArgument]
            assert option not in subFileMock.write.call_args_list[0][0][0]
        for option in optionsExpected:
            assert option in subFileMock.write.call_args_list[0][0][0]


@pytest.mark.parametrize(
    "localSchedd, expected", [(False, "-pool condorce.cern.ch:9619 -name condorce.cern.ch"), (True, "")]
)
def test_reset(setUp, localSchedd, expected):
    ceParameters = setUp

    htce = HTCE.HTCondorCEComputingElement(12345)
    htce.ceParameters = ceParameters
    htce.useLocalSchedd = True
    ceName = "condorce.cern.ch"
    htce.ceName = ceName
    htce._reset()
    assert htce.remoteScheddOptions == ""


@pytest.mark.parametrize(
    "localSchedd, expected",
    [
        (False, "condor_submit -terse -pool condorce.cern.ch:9619 -remote condorce.cern.ch dirac_pilot"),
        (True, "condor_submit -terse dirac_pilot"),
    ],
)
def test_submitJob(setUp, mocker, localSchedd, expected):
    ceParameters = setUp
    htce = HTCE.HTCondorCEComputingElement(12345)
    htce.ceParameters = ceParameters
    htce.useLocalSchedd = localSchedd
    ceName = "condorce.cern.ch"
    htce.ceName = ceName

    execMock = mocker.patch(MODNAME + ".executeGridCommand", return_value=S_OK((0, "123.0 - 123.0")))
    writeSubMock = mocker.patch(
        MODNAME + ".HTCondorCEComputingElement._HTCondorCEComputingElement__writeSub", return_value="dirac_pilot"
    )
    osMock = mocker.patch(MODNAME + ".os")

    with execMock, writeSubMock, osMock:
        result = htce.submitJob("pilot", "proxy", 1)

    assert result["OK"] is True
    assert " ".join(execMock.call_args_list[0][0][1]) == expected


@pytest.mark.parametrize(
    "jobIDList, jobID, ret, success, local",
    [
        ([], "", 0, True, True),
        ("", "", 0, True, True),
        (["htcondorce://condorce.foo.arg/123.0:::abc321"], "123.0", 0, True, True),
        ("htcondorce://condorce.foo.arg/123.0:::abc321", "123.0", 0, True, True),
        ("htcondorce://condorce.foo.arg/123.0:::abc321", "123.0", 1, False, True),
        (["htcondorce://condorce.foo.arg/333.3"], "333.3", 0, True, True),
        ("htcondorce://condorce.foo.arg/333.3", "333.3", 0, True, False),
    ],
)
def test_killJob(setUp, mocker, jobIDList, jobID, ret, success, local):
    ceParameters = setUp
    htce = HTCE.HTCondorCEComputingElement(12345)
    htce.ceName = "condorce.foo.arg"
    htce.useLocalSchedd = local
    htce.ceParameters = ceParameters
    htce._reset()

    commandsMock = mocker.patch(MODNAME + ".commands.getstatusoutput", return_value=(ret, ""))
    with commandsMock:
        ret = htce.killJob(jobIDList=jobIDList)

    assert ret["OK"] == success
    if jobID:
        commandsMock.assert_called_with("condor_rm %s %s" % (htce.remoteScheddOptions, jobID))



""" SiteInspectorAgent

  This agent inspect Sites, and evaluates policies that apply.

The following options can be set for the SiteInspectorAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN SiteInspectorAgent
  :end-before: ##END
  :dedent: 2
  :caption: SiteInspectorAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import concurrent.futures

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ResourceStatusSystem.PolicySystem.PEP import PEP

AGENT_NAME = "ResourceStatus/SiteInspectorAgent"


class SiteInspectorAgent(AgentModule):
    """SiteInspectorAgent

    The SiteInspectorAgent agent is an agent that is used to get the all the site names
    and trigger PEP to evaluate their status.
    """

    # Max number of worker threads by default
    __maxNumberOfThreads = 15

    # Inspection freqs, defaults, the lower, the higher priority to be checked.
    # Error state usually means there is a glitch somewhere, so it has the highest
    # priority.
    __checkingFreqs = {"Active": 20, "Degraded": 20, "Probing": 20, "Banned": 15, "Unknown": 10, "Error": 5}

    def __init__(self, *args, **kwargs):

        AgentModule.__init__(self, *args, **kwargs)

        self.rsClient = None
        self.clients = {}

    def initialize(self):
        """Standard initialize."""

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceManagementClient class: %s" % res["Message"])
            return res
        rmClass = res["Value"]

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceStatusClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceStatusClient class: %s" % res["Message"])
            return res
        rsClass = res["Value"]

        self.rsClient = rsClass()
        self.clients["ResourceStatusClient"] = rsClass()
        self.clients["ResourceManagementClient"] = rmClass()

        maxNumberOfThreads = self.am_getOption("maxNumberOfThreads", 15)
        self.log.info("Multithreaded with %d threads" % maxNumberOfThreads)
        self.threadPoolExecutor = concurrent.futures.ThreadPoolExecutor(max_workers=maxNumberOfThreads)

        return S_OK()

    def execute(self):
        """
        It gets the sites from the Database which are eligible to be re-checked.
        """

        utcnow = datetime.datetime.utcnow().replace(microsecond=0)
        future_to_element = {}

        # get the current status
        res = self.rsClient.selectStatusElement("Site", "Status")
        if not res["OK"]:
            return res

        # filter elements
        for site in res["Value"]:

            # Maybe an overkill, but this way I have NEVER again to worry about order
            # of elements returned by mySQL on tuples
            siteDict = dict(zip(res["Columns"], site))

            # This if-clause skips all the elements that should not be checked yet
            timeToNextCheck = self.__checkingFreqs[siteDict["Status"]]
            if utcnow <= siteDict["LastCheckTime"] + datetime.timedelta(minutes=timeToNextCheck):
                continue

            # We skip the elements with token different than "rs_svc"
            if siteDict["TokenOwner"] != "rs_svc":
                self.log.verbose("Skipping %s with token %s" % (siteDict["Name"], siteDict["TokenOwner"]))
                continue

            # if we are here, we process the current element
            self.log.verbose('"%s" # %s # %s' % (siteDict["Name"], siteDict["Status"], siteDict["LastCheckTime"]))

            lowerElementDict = {"element": "Site"}
            for key, value in siteDict.items():
                if len(key) >= 2:  # VO !
                    lowerElementDict[key[0].lower() + key[1:]] = value
            # We process lowerElementDict
            future = self.threadPoolExecutor.submit(self._execute, lowerElementDict)
            future_to_element[future] = siteDict["Name"]

        for future in concurrent.futures.as_completed(future_to_element):
            transID = future_to_element[future]
            try:
                future.result()
            except Exception as exc:
                self.log.exception("%s generated an exception: %s" % (transID, exc))
            else:
                self.log.info("Processed", transID)

        return S_OK()

    def _execute(self, site):
        """
        Method run by each of the thread that is in the ThreadPool.
        It evaluates the policies for such site and enforces the necessary actions.
        """

        pep = PEP(clients=self.clients)

        self.log.verbose(
            "%s ( VO=%s / status=%s / statusType=%s ) being processed"
            % (site["name"], site["vO"], site["status"], site["statusType"])
        )

        try:
            res = pep.enforce(site)
        except Exception:
            self.log.exception("Exception during enforcement")
            res = S_ERROR("Exception during enforcement")
        if not res["OK"]:
            self.log.error("Failed policy enforcement", res["Message"])
            return res

        resEnforce = res["Value"]

        oldStatus = resEnforce["decisionParams"]["status"]
        statusType = resEnforce["decisionParams"]["statusType"]
        newStatus = resEnforce["policyCombinedResult"]["Status"]
        reason = resEnforce["policyCombinedResult"]["Reason"]

        if oldStatus != newStatus:
            self.log.info(
                "%s (%s) is now %s ( %s ), before %s" % (site["name"], statusType, newStatus, reason, oldStatus)
            )

    def finalize(self):
        """graceful finalization"""

        self.log.info("Wait for threads to get empty before terminating the agent")
        self.threadPoolExecutor.shutdown()
        self.log.info("Threads are empty, terminating the agent...")
        return S_OK()



"""CSShellCmd class emulates the behaviour of a shell to edit the CS config.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import cmd
import os

import six

from DIRAC.Core.Base.CLI import CLI, colorize
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.Modificator import Modificator
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient


class CSShellCLI(CLI):
    def __init__(self):
        CLI.__init__(self)
        self.serverURL = ""
        self.serverName = ""
        self.modificator = None
        self.connected = False
        self.dirty = False
        self.root = "/"

        self.do_connect("")

    def update_prompt(self):
        if self.connected:
            self.prompt = "[" + colorize(self.serverName, "green") + ":" + self.root + " ]% "
        else:
            self.prompt = "[" + colorize("disconnected", "red") + ":" + self.root + " ]% "

    def do_connect(self, line):
        """connect
        Connect to the CS
        Usage: connect <URL> (Connect to the CS at the specified URL)
               connect       (Connect to the default CS URL of your config)
        """
        if line == "":
            self.serverURL = gConfigurationData.getMasterServer()
            self.serverName = gConfigurationData.getName()
        else:
            self.serverURL = self.serverName = line

        if self.serverURL is None:
            print("Unable to connect to the default server. Maybe you don't have a proxy ?")
            return self.do_disconnect("")

        print("Trying to connect to " + self.serverURL + "...", end=" ")

        self.modificator = Modificator(ConfigurationClient(url=self.serverURL))
        rv = self.modificator.loadFromRemote()
        rv2 = self.modificator.loadCredentials()

        if rv["OK"] == False or rv2["OK"] == False:
            print("failed: ", end=" ")
            if rv["OK"] is False:
                print(rv["Message"])
            else:
                print(rv2["Message"])
            self.connected = False
            self.update_prompt()
        else:
            self.connected = True
            self.update_prompt()
            print("done.")

    def do_disconnect(self, _line):
        """Disconnect from CS"""
        if self.connected and self.dirty:
            res = six.moves.input("Do you want to commit your changes into the CS ? [y/N] ")
            if res.lower() in ["y", "yes"]:
                self.do_commit("")

        self.serverURL = self.serverName = ""
        self.modificator = None
        self.connected = False
        self.update_prompt()

    def do_ls(self, line):
        """ls
        List the sections and options of CS of the current root"""
        if self.connected:
            secs = self.modificator.getSections(self.root)
            opts = self.modificator.getOptions(self.root)
            if line.startswith("-") and "l" in line:
                for i in secs:
                    print(colorize(i, "blue") + "  ")
                for i in opts:
                    print(i + " ")
            else:
                for i in secs:
                    print(colorize(i, "blue") + "  ", end=" ")
                for i in opts:
                    print(i + " ", end=" ")
                print("")

    def do_cd(self, line):
        """cd
        Go one directory deeper in the CS"""
        # Check if invariant holds
        if self.connected:
            assert self.root == "/" or not self.root.endswith("/")
            assert self.root.startswith("/")
            secs = self.modificator.getSections(self.root)
            if line == "..":
                self.root = os.path.dirname(self.root)
                self.update_prompt()
            else:
                if os.path.normpath(line) in secs:
                    if self.root == "/":
                        self.root = self.root + os.path.normpath(line)
                    else:
                        self.root = self.root + "/" + os.path.normpath(line)
                    self.update_prompt()
                else:
                    print("cd: no such section: " + line)

    def complete_cd(self, text, _line, _begidx, _endidx):
        secs = self.modificator.getSections(self.root)
        return [(s + "/") for s in secs if s.startswith(text)]

    def do_cat(self, line):
        """cat
        Read the content of an option in the CS"""
        if self.connected:
            opts = self.modificator.getOptionsDict(self.root)

            if line in opts:
                print(opts[line])
            else:
                print("cat: No such option")

    def complete_cat(self, text, _line, _begidx, _endidx):
        opts = self.modificator.getOptions(self.root)
        return [o for o in opts if o.startswith(text)]

    do_less = do_cat
    complete_less = complete_cat

    def do_mkdir(self, line):
        """mkdir
        Create a new section in the CS"""
        if self.connected:
            self.modificator.createSection(self.root + "/" + line)
            self.dirty = True

    complete_mkdir = complete_cd

    def do_rmdir(self, line):
        """rmdir
        Delete a section in the CS"""
        if self.connected:
            self.modificator.removeSection(self.root + "/" + line)
            self.dirty = True

    complete_rmdir = complete_cd

    def do_rm(self, line):
        """rm
        Delete an option in the CS"""
        if self.connected:
            self.modificator.removeOption(self.root + "/" + line)
            self.dirty = True

    complete_rm = complete_cat

    def do_set(self, line):
        """set
        Set an option in the CS (or create it if it does not exists)
        Usage: set <str> to set a string option (will be stored as a string in CS)
               set <str>,<str>,... to set a list option (will be stored as a list in CS)
        """
        if self.connected:
            line = line.split(" ", 2)
            if len(line) != 2:
                print("Usage: set <key> <value>")
            else:
                self.modificator.setOptionValue(self.root + "/" + line[0], line[1])
                self.dirty = True

    complete_set = complete_cat

    def do_unset(self, line):
        """unset
        Unset an option in the CS: Making the option equal to the
        empty string."""
        if self.connected:
            self.modificator.setOptionValue(self.root + "/" + line, "")
            self.dirty = True

    complete_unset = complete_cat

    def do_commit(self, _line):
        """commit
        Commit the modifications to the CS"""
        if self.connected and self.dirty:
            self.modificator.commit()

    def default(self, line):
        """Override [Cmd.default(line)] function."""
        if line == "EOF":
            if self.prompt:
                print()
            return self.do_quit(line)
        else:
            cmd.Cmd.default(self, line)

    def do_quit(self, _line):
        """quit
        Quit"""
        self.do_disconnect("")
        CLI.do_quit(self, _line)



########################################################################
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2011/01/17 08:17:58
########################################################################

""".. module:: ListTestCase

Test cases for DIRAC.Core.Utilities.List module.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

# sut
from DIRAC.Core.Utilities import List

__RCSID__ = "$Id$"


########################################################################
class ListTestCase(unittest.TestCase):
    """py:class ListTestCase
    Test case for DIRAC.Core.Utilities.List module.
    """

    def testUniqueElements(self):
        """uniqueElements tests"""
        # empty list
        aList = []
        self.assertEqual(List.uniqueElements(aList), [])
        # redundant elements
        aList = [1, 1, 2, 3]
        self.assertEqual(List.uniqueElements(aList), [1, 2, 3])

    def testAppendUnique(self):
        """appendUnique tests"""
        # empty
        aList = []
        List.appendUnique(aList, None)
        self.assertEqual(aList, [None])
        # redundant element
        aList = [1, 2, 3]
        List.appendUnique(aList, 1)
        self.assertEqual(aList, [1, 2, 3])
        # all unique
        aList = [1, 2]
        List.appendUnique(aList, 3)
        self.assertEqual(aList, [1, 2, 3])

    def testRandomize(self):
        """randomize tests"""
        # empty list
        aList = []
        randList = List.randomize(aList)
        self.assertEqual(randList, [])
        # non empty
        aList = ["1", "2", "3"]
        randList = List.randomize(aList)
        self.assertEqual(len(aList), len(randList))
        for x in aList:
            self.assertEqual(x in randList, True)
        for x in randList:
            self.assertEqual(x in aList, True)

    def testPop(self):
        """pop tests"""
        # empty list
        aList = []
        x = List.pop(aList, 1)
        self.assertEqual(aList, [])
        self.assertEqual(x, None)
        # pop
        aList = [1, 2, 3]
        x = List.pop(aList, 2)
        self.assertEqual(x, 2)
        self.assertEqual(aList, [1, 3])

    def testStringListToString(self):
        """stringListToString tests"""
        # empty list
        aList = []
        aStr = List.stringListToString(aList)
        self.assertEqual(aStr, "")
        # not string elements (should it raise an exception???)
        aList = ["a", 1]
        aStr = List.stringListToString(aList)
        self.assertEqual(aStr, "'a','1'")
        # normal list
        aList = ["a", "b", "c"]
        aStr = List.stringListToString(aList)
        self.assertEqual(aStr, "'a','b','c'")

    def testIntListToString(self):
        """intListToString"""
        # empty list
        aList = []
        aStr = List.intListToString(aList)
        self.assertEqual(aStr, "")
        # int list
        aList = [1, 2, 3]
        aStr = List.intListToString(aList)
        self.assertEqual(aStr, "1,2,3")
        # mixture elements (should it raise an exception???)
        aList = ["1", 2, 3]
        aStr = List.intListToString(aList)
        self.assertEqual(aStr, "1,2,3")

    def testFromChar(self):
        """fromChar tests"""
        # empty string
        aStr = ""
        self.assertEqual(List.fromChar(aStr, "-"), [])
        # wrong sep (should it raise an exception???)
        aStr = "a:b:c"
        self.assertEqual(List.fromChar(aStr, "-"), ["a:b:c"])
        # norman behavior
        aStr = "a:b:c"
        self.assertEqual(List.fromChar(aStr, ":"), ["a", "b", "c"])
        # only sep
        aStr = ","
        self.assertEqual(List.fromChar(aStr, ","), [])
        # too many separators
        aStr = "a,,b,,c,,,"
        self.assertEqual(List.fromChar(aStr, ","), ["a", "b", "c"])

    def testBreakListIntoChunks(self):
        """breakListIntoChunks tests"""
        # empty list
        aList = []
        self.assertEqual(List.breakListIntoChunks(aList, 5), [])
        # negative number of chunks
        try:
            List.breakListIntoChunks([], -2)
        except Exception as val:
            self.assertEqual(isinstance(val, RuntimeError), True)
            self.assertEqual(str(val), "chunkSize cannot be less than 1")

        # normal behavior
        aList = list(range(10))
        self.assertEqual(List.breakListIntoChunks(aList, 5), [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
        # and once again this time with a rest
        aList = list(range(10))
        self.assertEqual(List.breakListIntoChunks(aList, 4), [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9]])
        # almost empty list, too many chunks
        aList = [1]
        self.assertEqual(List.breakListIntoChunks(aList, 2), [[1]])


# test suite execution
if __name__ == "__main__":
    TESTLOADER = unittest.TestLoader()
    SUITE = TESTLOADER.loadTestsFromTestCase(ListTestCase)
    unittest.TextTestRunner(verbosity=3).run(SUITE)



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-get-normalized-queue-length.py
# Author :  Ricardo Graciani
########################################################################
"""
Report Normalized CPU length of queue

This script was used by the dirac-pilot script to set the CPUTime limit for
the matching but now this is no more the case.

Example:
  $ dirac-wms-get-normalized-queue-length  cclcgceli03.in2p3.fr:2119/jobmanager-bqs-long
  cclcgceli03.in2p3.fr:2119/jobmanager-bqs-long 857400.0
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.WorkloadManagementSystem.Client.CPUNormalization import queueNormalizedCPU


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Queue: GlueCEUniqueID of the Queue (ie, juk.nikhef.nl:8443/cream-pbs-lhcb)"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    exitCode = 0

    for ceUniqueID in args:

        normCPU = queueNormalizedCPU(ceUniqueID)

        if not normCPU["OK"]:
            print("ERROR %s:" % ceUniqueID, normCPU["Message"])
            exitCode = 2
            continue
        print(ceUniqueID, normCPU["Value"])

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Collection of utilities for dealing with security files (i.e. proxy files)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import stat
import tempfile

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security.Locations import getProxyLocation


def writeToProxyFile(proxyContents, fileName=False):
    """Write a proxy string to file

    arguments:
      - proxyContents : string object to dump to file
      - fileName : filename to dump to
    """
    if not fileName:
        try:
            fd, proxyLocation = tempfile.mkstemp()
            os.close(fd)
        except IOError:
            return S_ERROR(DErrno.ECTMPF)
        fileName = proxyLocation
    try:
        with open(fileName, "wb") as fd:
            fd.write(proxyContents)
    except Exception as e:
        return S_ERROR(DErrno.EWF, " %s: %s" % (fileName, repr(e).replace(",)", ")")))
    try:
        os.chmod(fileName, stat.S_IRUSR | stat.S_IWUSR)
    except Exception as e:
        return S_ERROR(DErrno.ESPF, "%s: %s" % (fileName, repr(e).replace(",)", ")")))
    return S_OK(fileName)


def writeChainToProxyFile(proxyChain, fileName):
    """
    Write an X509Chain to file

    arguments:
      - proxyChain : X509Chain object to dump to file
      - fileName : filename to dump to
    """
    retVal = proxyChain.dumpAllToString()
    if not retVal["OK"]:
        return retVal
    return writeToProxyFile(retVal["Value"], fileName)


def writeChainToTemporaryFile(proxyChain):
    """
    Write a proxy chain to a temporary file
    return S_OK( string with name of file )/ S_ERROR
    """
    try:
        fd, proxyLocation = tempfile.mkstemp()
        os.close(fd)
    except IOError:
        return S_ERROR(DErrno.ECTMPF)
    retVal = writeChainToProxyFile(proxyChain, proxyLocation)
    if not retVal["OK"]:
        try:
            os.unlink(proxyLocation)
        except Exception:
            pass
        return retVal
    return S_OK(proxyLocation)


def deleteMultiProxy(multiProxyDict):
    """
    Delete a file from a multiProxyArgument if needed
    """
    if multiProxyDict["tempFile"]:
        try:
            os.unlink(multiProxyDict["file"])
        except Exception:
            pass


def multiProxyArgument(proxy=False):
    """
    Load a proxy:


    :param proxy: param can be:

        * Default -> use current proxy
        * string -> upload file specified as proxy
        * X509Chain -> use chain

    :returns:  S_OK/S_ERROR

      .. code-block:: python

          S_OK( { 'file' : <string with file location>,
                  'chain' : X509Chain object,
                  'tempFile' : <True if file is temporal>
                } )
          S_ERROR

    """
    tempFile = False
    # Set env
    if isinstance(proxy, X509Chain):
        tempFile = True
        retVal = writeChainToTemporaryFile(proxy)
        if not retVal["OK"]:
            return retVal
        proxyLoc = retVal["Value"]
    else:
        if not proxy:
            proxyLoc = getProxyLocation()
            if not proxyLoc:
                return S_ERROR(DErrno.EPROXYFIND)
        if isinstance(proxy, six.string_types):
            proxyLoc = proxy
        # Load proxy
        proxy = X509Chain()
        retVal = proxy.loadProxyFromFile(proxyLoc)
        if not retVal["OK"]:
            return S_ERROR(DErrno.EPROXYREAD, "ProxyLocation: %s" % proxyLoc)
    return S_OK({"file": proxyLoc, "chain": proxy, "tempFile": tempFile})



"""
ReportCLI class implementing command line interface to DIRAC Accounting
ReportGenerator Service. It is not complete yet

Once ready it could be used with a script as simple as:


from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

Script.localCfg.addDefaultEntry("LogLevel", "info")
Script.parseCommandLine()

from DIRAC.AccountingSystem.Client.ReportCLI import ReportCLI

if __name__=="__main__":
    reli = ReportCLI()
    reli.start()


"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys
import datetime

from DIRAC.Core.Base.CLI import CLI, colorize
from DIRAC.AccountingSystem.Client.ReportsClient import ReportsClient
from DIRAC import gLogger


class ReportCLI(CLI):
    def __init__(self):
        CLI.__init__(self)
        self.do_connect(None)

    def start(self):
        """
        Start the command loop
        """
        if not self.connected:
            gLogger.error("Client is not connected")
        try:
            self.cmdloop()
        except KeyboardInterrupt as v:
            gLogger.warn("Received a keyboard interrupt.")
            self.do_quit("")

    def do_connect(self, args):
        """
        Tries to connect to the server
            Usage: connect
        """
        gLogger.info("Trying to connect to server")
        self.connected = False
        self.prompt = "(%s)> " % colorize("Not connected", "red")
        retVal = ReportsClient().ping()
        if retVal["OK"]:
            self.prompt = "(%s)> " % colorize("Connected", "green")
            self.connected = True

    def printComment(self, comment):
        commentList = comment.split("\n")
        for commentLine in commentList[:-1]:
            print("# %s" % commentLine.strip())

    def showTraceback(self):
        import traceback

        type, value = sys.exc_info()[:2]
        print("________________________\n")
        print("Exception", type, ":", value)
        traceback.print_tb(sys.exc_info()[2])
        print("________________________\n")

    def __getDatetimeFromArg(self, dtString):
        if len(dtString) != 12:
            return False
        dt = datetime.datetime(year=int(dtString[0:4]), month=int(dtString[4:6]), day=int(dtString[6:8]))
        dt += datetime.timedelta(hours=int(dtString[8:10]), minutes=int(dtString[10:12]))
        return dt



""" This helper looks in the /Operations section of the CS, considering its specific nature:
    the /Operations section is designed in a way that each configuration can be specific to a Setup,
    while maintaining a default.

    So, for example, given the following /Operations section::

      Operations/
          Defaults/
              someSection/
                  someOption = someValue
                  aSecondOption = aSecondValue
          Production/
              someSection/
                  someOption = someValueInProduction
                  aSecondOption = aSecondValueInProduction
          Certification/
              someSection/
                  someOption = someValueInCertification

    The following calls would give different results based on the setup::

      Operations().getValue('someSection/someOption')
        - someValueInProduction if we are in 'Production' setup
        - someValueInCertification if we are in 'Certification' setup

      Operations().getValue('someSection/aSecondOption')
        - aSecondValueInProduction if we are in 'Production' setup
        - aSecondValue if we are in 'Certification' setup    <- looking in Defaults
                                                                since there's no Certification/someSection/aSecondOption


    At the same time, for multi-VO installations, it is also possible to specify different options per-VO,
    like the following::

      Operations/
          aVOName/
              Defaults/
                  someSection/
                      someOption = someValue
                      aSecondOption = aSecondValue
              Production/
                  someSection/
                      someOption = someValueInProduction
                      aSecondOption = aSecondValueInProduction
              Certification/
                  someSection/
                      someOption = someValueInCertification
          anotherVOName/
              Defaults/
                  someSectionName/
                      someOptionX = someValueX
                      aSecondOption = aSecondValue
              setupName/
                  someSection/
                      someOption = someValueInProduction
                      aSecondOption = aSecondValueInProduction

    For this case it becomes then important for the Operations() objects to know the VO name
    for which we want the information, and this can be done in the following ways.

    1. by specifying the VO name directly::

         Operations(vo=anotherVOName).getValue('someSectionName/someOptionX')

    2. by give a group name::

         Operations(group=thisIsAGroupOfVO_X).getValue('someSectionName/someOptionX')

    3. if no VO nor group is provided, the VO will be guessed from the proxy,
    but this works iff the object is instantiated by a proxy (and not, e.g., using a server certificate)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from six.moves import _thread as thread
import os
from diraccfg import CFG
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers import Registry, CSGlobals
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
from DIRAC.Core.Utilities import LockRing
from DIRAC.Core.Utilities.DErrno import ESECTION


class Operations(object):
    """Operations class

    The /Operations CFG section is maintained in a cache by an Operations object
    """

    __cache = {}
    __cacheVersion = 0
    __cacheLock = LockRing.LockRing().getLock()

    def __init__(self, vo=False, group=False, setup=False):
        """c'tor

        Setting some defaults
        """
        self.__uVO = vo
        self.__uGroup = group
        self.__uSetup = setup
        self.__vo = False
        self.__setup = False
        self.__discoverSettings()

    def __discoverSettings(self):
        """Discovers the vo and the setup"""
        # Set the VO
        globalVO = CSGlobals.getVO()
        if globalVO:
            self.__vo = globalVO
        elif self.__uVO:
            self.__vo = self.__uVO
        elif self.__uGroup:
            self.__vo = Registry.getVOForGroup(self.__uGroup)
            if not self.__vo:
                self.__vo = False
        else:
            result = getVOfromProxyGroup()
            if result["OK"]:
                self.__vo = result["Value"]
        # Set the setup
        self.__setup = False
        if self.__uSetup:
            self.__setup = self.__uSetup
        else:
            self.__setup = CSGlobals.getSetup()

    def __getCache(self):
        Operations.__cacheLock.acquire()
        try:
            currentVersion = gConfigurationData.getVersion()
            if currentVersion != Operations.__cacheVersion:
                Operations.__cache = {}
                Operations.__cacheVersion = currentVersion

            cacheKey = (self.__vo, self.__setup)
            if cacheKey in Operations.__cache:
                return Operations.__cache[cacheKey]

            mergedCFG = CFG()

            for path in self.__getSearchPaths():
                pathCFG = gConfigurationData.mergedCFG[path]
                if pathCFG:
                    mergedCFG = mergedCFG.mergeWith(pathCFG)

            Operations.__cache[cacheKey] = mergedCFG

            return Operations.__cache[cacheKey]
        finally:
            try:
                Operations.__cacheLock.release()
            except thread.error:
                pass

    def __getSearchPaths(self):
        paths = ["/Operations/Defaults", "/Operations/%s" % self.__setup]
        if not self.__vo:
            globalVO = CSGlobals.getVO()
            if not globalVO:
                return paths
            self.__vo = CSGlobals.getVO()
        paths.append("/Operations/%s/Defaults" % self.__vo)
        paths.append("/Operations/%s/%s" % (self.__vo, self.__setup))
        return paths

    def getValue(self, optionPath, defaultValue=None):
        return self.__getCache().getOption(optionPath, defaultValue)

    def __getCFG(self, sectionPath):
        cacheCFG = self.__getCache()
        section = cacheCFG.getRecursive(sectionPath)
        if not section:
            return S_ERROR(ESECTION, "%s in Operations does not exist" % sectionPath)
        sectionCFG = section["value"]
        if isinstance(sectionCFG, six.string_types):
            return S_ERROR("%s in Operations is not a section" % sectionPath)
        return S_OK(sectionCFG)

    def getSections(self, sectionPath, listOrdered=False):
        result = self.__getCFG(sectionPath)
        if not result["OK"]:
            return result
        sectionCFG = result["Value"]
        return S_OK(sectionCFG.listSections(listOrdered))

    def getOptions(self, sectionPath, listOrdered=False):
        result = self.__getCFG(sectionPath)
        if not result["OK"]:
            return result
        sectionCFG = result["Value"]
        return S_OK(sectionCFG.listOptions(listOrdered))

    def getOptionsDict(self, sectionPath):
        result = self.__getCFG(sectionPath)
        if not result["OK"]:
            return result
        sectionCFG = result["Value"]
        data = {}
        for opName in sectionCFG.listOptions():
            data[opName] = sectionCFG[opName]
        return S_OK(data)

    def getPath(self, option, vo=False, setup=False):
        """
        Generate the CS path for an option:

          - if vo is not defined, the helper's vo will be used for multi VO installations
          - if setup evaluates False (except None) -> The helpers setup will  be used
          - if setup is defined -> whatever is defined will be used as setup
          - if setup is None -> Defaults will be used

        :param option: path with respect to the Operations standard path
        :type option: string
        """

        for path in self.__getSearchPaths():
            optionPath = os.path.join(path, option)
            value = gConfig.getValue(optionPath, "NoValue")
            if value != "NoValue":
                return optionPath
        return ""



#!/usr/bin/env python
"""
Print list of users with proxies.

Example:
  $ dirac-admin-users-with-proxy
  * vhamar
  DN         : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
  group      : dirac_admin
  not after  : 2011-06-29 12:04:25
  persistent : False
  -
  DN         : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
  group      : dirac_pilot
  not after  : 2011-06-29 12:04:27
  persistent : False
  -
  DN         : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
  group      : dirac_user
  not after  : 2011-06-29 12:04:30
  persistent : True
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import DIRAC
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager

__RCSID__ = "$Id$"


class Params(object):

    limited = False
    proxyPath = False
    proxyLifeTime = 3600

    def setProxyLifeTime(self, arg):
        try:
            fields = [f.strip() for f in arg.split(":")]
            self.proxyLifeTime = int(fields[0]) * 3600 + int(fields[1]) * 60
        except Exception:
            print("Can't parse %s time! Is it a HH:MM?" % arg)
            return DIRAC.S_ERROR("Can't parse time argument")
        return DIRAC.S_OK()

    def registerCLISwitches(self):
        Script.registerSwitch("v:", "valid=", "Required HH:MM for the users", self.setProxyLifeTime)


@Script()
def main():
    params = Params()
    params.registerCLISwitches()
    Script.parseCommandLine(ignoreErrors=True)
    result = gProxyManager.getDBContents()
    if not result["OK"]:
        print("Can't retrieve list of users: %s" % result["Message"])
        DIRAC.exit(1)

    keys = result["Value"]["ParameterNames"]
    records = result["Value"]["Records"]
    dataDict = {}
    now = Time.dateTime()
    for record in records:
        expirationDate = record[3]
        dt = expirationDate - now
        secsLeft = dt.days * 86400 + dt.seconds
        if secsLeft > params.proxyLifeTime:
            userName, userDN, userGroup, _, persistent = record
            if userName not in dataDict:
                dataDict[userName] = []
            dataDict[userName].append((userDN, userGroup, expirationDate, persistent))

    for userName in dataDict:
        print("* %s" % userName)
        for iP in range(len(dataDict[userName])):
            data = dataDict[userName][iP]
            print(" DN         : %s" % data[0])
            print(" group      : %s" % data[1])
            print(" not after  : %s" % Time.toString(data[2]))
            print(" persistent : %s" % data[3])
            if iP < len(dataDict[userName]) - 1:
                print(" -")

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" Test class for Pilot Status Agent
"""

# imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import pytest
from mock import MagicMock

# DIRAC Components
from DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent import PilotStatusAgent
from DIRAC import gLogger, S_OK

# Mock objects
mockReply = MagicMock()
mockReply.return_value = {"OK": True, "Value": []}
mockAM = MagicMock()
mockNone = MagicMock()
mockNone.return_value = None
mockOK = MagicMock()
mockOK.return_value = {"OK": False}

gLogger.setLevel("DEBUG")


@pytest.fixture
def psa(mocker):
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.AgentModule.__init__")
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )
    module_str = "DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.PilotAgentsDB.selectPilots"
    mocker.patch(module_str, side_effect=mockReply)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.PilotAgentsDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.JobDB.__init__", side_effect=mockNone)
    module_str = "DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.PilotAgentsDB.buildCondition"
    mocker.patch(module_str, side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.PilotAgentsDB._query", side_effect=mockOK)
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent.PilotAgentsDB._escapeString",
        lambda s, c: S_OK('"%s"' % s),
    )  # To bypass "connection.escape_string"

    psa = PilotStatusAgent()
    psa._AgentModule__configDefaults = mockAM
    psa.initialize()
    psa.log = gLogger
    psa.pilotDB.log = gLogger
    psa.pilotDB.logger = gLogger
    psa.pilotStalledDays = 3

    return psa


@pytest.mark.parametrize(
    "mockReplyInput, expected",
    [
        ({"OK": True, "Value": False}, {"OK": True, "Value": None}),
        (
            {"OK": True, "Value": ["Test"]},
            {
                "OK": False,
                "Message": "No pilots found for PilotJobReference(s): ['Test']",
            },
        ),
        ({"OK": False, "Message": "Test"}, {"OK": False, "Message": "Test"}),
        ({"OK": True, "Value": []}, {"OK": True, "Value": None}),
    ],
)
def test_handleOldPilots(psa, mockReplyInput, expected):
    """Testing PilotStatusAgent().handleOldPilots()"""

    mockReply.return_value = mockReplyInput
    connection = "Test"
    result = psa.handleOldPilots(connection)
    assert result["OK"] == expected["OK"]
    if result["OK"]:
        assert result["Value"] == expected["Value"]
    else:
        if "Message" in result:
            assert result["Message"] == expected["Message"]



#!/usr/bin/env python
########################################################################
# File :    dirac-proxy-init.py
# Author :  Adrian Casajus
########################################################################

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys
import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


class Params:
    proxyLoc = False
    dnAsUsername = False

    def setProxyLocation(self, arg):
        self.proxyLoc = arg
        return DIRAC.S_OK()

    def setDNAsUsername(self, arg):
        self.dnAsUsername = True
        return DIRAC.S_OK()

    def showVersion(self, arg):
        print("Version:")
        print(" ", __RCSID__)
        sys.exit(0)
        return DIRAC.S_OK()


@Script()
def main():
    params = Params()

    Script.registerSwitch("f:", "file=", "File to use as proxy", params.setProxyLocation)
    Script.registerSwitch("D", "DN", "Use DN as myproxy username", params.setDNAsUsername)
    Script.registerSwitch("i", "version", "Print version", params.showVersion)

    Script.addDefaultOptionValue("LogLevel", "always")
    Script.parseCommandLine()

    from DIRAC.Core.Security.MyProxy import MyProxy
    from DIRAC.Core.Security import Locations

    if not params.proxyLoc:
        params.proxyLoc = Locations.getProxyLocation()

    if not params.proxyLoc:
        print("Can't find any valid proxy")
        sys.exit(1)
    print("Uploading proxy file %s" % params.proxyLoc)

    mp = MyProxy()
    retVal = mp.uploadProxy(params.proxyLoc, params.dnAsUsername)
    if not retVal["OK"]:
        print("Can't upload proxy:")
        print(" ", retVal["Message"])
        sys.exit(1)
    print("Proxy uploaded")
    sys.exit(0)


if __name__ == "__main__":
    main()
